# coding: utf-8


import logging
import math
import os
import subprocess
import tempfile
import time

import numpy as np
from monty.dev import requires
from monty.json import jsanitize, MSONable
from monty.os import cd
from monty.os.path import which
from scipy import constants
from scipy.spatial import distance

from pymatgen.core.lattice import Lattice
from pymatgen.core.units import Energy, Length
from pymatgen.electronic_structure.bandstructure import \
    BandStructureSymmLine, Kpoint
from pymatgen.electronic_structure.core import Orbital
from pymatgen.electronic_structure.dos import Dos, Spin, CompleteDos
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.symmetry.bandstructure import HighSymmKpath


"""
This module provides classes to run and analyze boltztrap on pymatgen band
structure objects. Boltztrap is a software interpolating band structures and
computing materials properties from this band structure using Boltzmann
semi-classical transport theory.

Boltztrap has been developed by Georg Madsen.

http://www.icams.de/content/research/software-development/boltztrap/

You need version 1.2.3 or higher

References are::

    Madsen, G. K. H., and Singh, D. J. (2006).
    BoltzTraP. A code for calculating band-structure dependent quantities.
    Computer Physics Communications, 175, 67-71
"""

__author__ = "Geoffroy Hautier, Zachary Gibbs, Francesco Ricci, Anubhav Jain"
__copyright__ = "Copyright 2013, The Materials Project"
__version__ = "1.1"
__maintainer__ = "Geoffroy Hautier"
__email__ = "geoffroy@uclouvain.be"
__status__ = "Development"
__date__ = "August 23, 2013"


class BoltztrapRunner(MSONable):
    """
    This class is used to run Boltztrap on a band structure object.

    Args:
            bs:
                A band structure object
            nelec:
                the number of electrons
            dos_type:
                two options for the band structure integration: "HISTO"
                (histogram) or "TETRA" using the tetrahedon method. TETRA
                typically gives better results (especially for DOSes)
                but takes more time
            energy_grid:
                the energy steps used for the integration (eV)
            lpfac:
                the number of interpolation points in the real space. By
                default 10 gives 10 time more points in the real space than
                the number of kpoints given in reciprocal space
            run_type:
                type of boltztrap usage. by default
                - BOLTZ: (default) compute transport coefficients
                - BANDS: interpolate all bands contained in the energy range
                specified in energy_span_around_fermi variable, along specified
                k-points
                - DOS: compute total and partial dos (custom BoltzTraP code
                needed!)
                - FERMI: compute fermi surface or more correctly to
                get certain bands interpolated
            band_nb:
                indicates a band number. Used for Fermi Surface interpolation
                (run_type="FERMI")
            spin:
                specific spin component (1: up, -1: down) of the band selected
                in FERMI mode (mandatory).
            cond_band:
                if a conduction band is specified in FERMI mode,
                set this variable as True
            tauref:
                reference relaxation time. Only set to a value different than
                zero if we want to model beyond the constant relaxation time.
            tauexp:
                exponent for the energy in the non-constant relaxation time
                approach
            tauen:
                reference energy for the non-constant relaxation time approach
            soc:
                results from spin-orbit coupling (soc) computations give
                typically non-polarized (no spin up or down) results but single
                electron occupations. If the band structure comes from a soc
                computation, you should set soc to True (default False)
            doping:
                the fixed doping levels you want to compute. Boltztrap provides
                both transport values depending on electron chemical potential
                (fermi energy) and for a series of fixed carrier
                concentrations. By default, this is set to 1e16 to 1e22 in
                increments of factors of 10.
            energy_span_around_fermi:
                usually the interpolation is not needed on the entire energy
                range but on a specific range around the fermi level.
                This energy gives this range in eV. by default it is 1.5 eV.
                If DOS or BANDS type are selected, this range is automatically
                set to cover the entire energy range.
            scissor:
                scissor to apply to the band gap (eV). This applies a scissor
                operation moving the band edges without changing the band
                shape. This is useful to correct the often underestimated band
                gap in DFT. Default is 0.0 (no scissor)
            kpt_line:
                list of fractional coordinates of kpoints as arrays or list of
                Kpoint objects for BANDS mode calculation (standard path of
                high symmetry k-points is automatically set as default)
            tmax:
                Maximum temperature (K) for calculation (default=1300)
            tgrid:
                Temperature interval for calculation (default=50)
            symprec: 1e-3 is the default in pymatgen. If the kmesh has been
                generated using a different symprec, it has to be specified
                to avoid a "factorization error" in BoltzTraP calculation.
                If a kmesh that spans the whole Brillouin zone has been used,
                or to disable all the symmetries, set symprec to None.
            cb_cut: by default 10% of the highest conduction bands are 
                removed because they are often not accurate. 
                Tune cb_cut to change the percentage (0-100) of bands 
                that are removed.
            timeout: overall time limit (in seconds): mainly to avoid infinite
                loop when trying to find Fermi levels.
    """

    @requires(which('x_trans'),
              "BoltztrapRunner requires the executables 'x_trans' to be in "
              "the path. Please download the Boltztrap at http://"
              "www.icams.de/content/research/software-development/boltztrap/ "
              "and follow the instructions in the README to compile "
              "Bolztrap accordingly. Then add x_trans to your path")
    def __init__(self, bs, nelec, dos_type="HISTO", energy_grid=0.005,
                 lpfac=10, run_type="BOLTZ", band_nb=None, tauref=0, tauexp=0,
                 tauen=0, soc=False, doping=None, energy_span_around_fermi=1.5,
                 scissor=0.0, kpt_line=None, spin=None, cond_band=False,
                 tmax=1300, tgrid=50, symprec=1e-3, cb_cut=10, timeout=7200):
        self.lpfac = lpfac
        self._bs = bs
        self._nelec = nelec
        self.dos_type = dos_type
        self.energy_grid = energy_grid
        self.error = []
        self.run_type = run_type
        self.band_nb = band_nb
        self.spin = spin
        self.cond_band = cond_band
        self.tauref = tauref
        self.tauexp = tauexp
        self.tauen = tauen
        self.soc = soc
        self.kpt_line = kpt_line
        self.cb_cut = cb_cut/100.
        if isinstance(doping, list) and len(doping) > 0:
            self.doping = doping
        else:
            self.doping = []
            for d in [1e16, 1e17, 1e18, 1e19, 1e20, 1e21]:
                self.doping.extend([1 * d, 2.5 * d, 5 * d, 7.5 * d])
            self.doping.append(1e22)
        self.energy_span_around_fermi = energy_span_around_fermi
        self.scissor = scissor
        self.tmax = tmax
        self.tgrid = tgrid
        self._symprec = symprec
        if self.run_type in ("DOS", "BANDS"):
            self._auto_set_energy_range()
        self.timeout = timeout
        self.start_time = time.time()

    def _auto_set_energy_range(self):
        """
        automatically determine the energy range as min/max eigenvalue
        minus/plus the buffer_in_ev
        """
        emins = [min([e_k[0] for e_k in self._bs.bands[Spin.up]])]
        emaxs = [max([e_k[0] for e_k in self._bs.bands[Spin.up]])]

        if self._bs.is_spin_polarized:
            emins.append(min([e_k[0] for e_k in
                              self._bs.bands[Spin.down]]))

            emaxs.append(max([e_k[0] for e_k in
                              self._bs.bands[Spin.down]]))

        min_eigenval = Energy(min(emins) - self._bs.efermi, "eV"). \
            to("Ry")
        max_eigenval = Energy(max(emaxs) - self._bs.efermi, "eV"). \
            to("Ry")

        # set energy range to buffer around min/max EV
        # buffer does not increase CPU time but will help get equal
        # energies for spin up/down for band structure
        const = Energy(2, "eV").to("Ry")
        self._ll = min_eigenval - const
        self._hl = max_eigenval + const

        en_range = Energy(max((abs(self._ll), abs(self._hl))),
                          "Ry").to("eV")

        self.energy_span_around_fermi = en_range * 1.01
        print("energy_span_around_fermi = ",
              self.energy_span_around_fermi)

    @property
    def bs(self):
        return self._bs

    @property
    def nelec(self):
        return self._nelec

    def write_energy(self, output_file):
        with open(output_file, 'w') as f:
            f.write("test\n")
            f.write("{}\n".format(len(self._bs.kpoints)))

            if self.run_type == "FERMI":
                sign = -1.0 if self.cond_band else 1.0
                for i in range(len(self._bs.kpoints)):
                    eigs = []
                    eigs.append(Energy(
                        self._bs.bands[Spin(self.spin)][self.band_nb][i] -
                        self._bs.efermi, "eV").to("Ry"))
                    f.write("%12.8f %12.8f %12.8f %d\n"
                            % (self._bs.kpoints[i].frac_coords[0],
                               self._bs.kpoints[i].frac_coords[1],
                               self._bs.kpoints[i].frac_coords[2],
                               len(eigs)))
                    for j in range(len(eigs)):
                        f.write("%18.8f\n" % (sign * float(eigs[j])))

            else:
                for i, kpt in enumerate(self._bs.kpoints):
                    eigs = []
                    if self.run_type == "DOS":
                        spin_lst = [self.spin]
                    else:
                        spin_lst = self._bs.bands

                    for spin in spin_lst:
                        # use 90% of bottom bands since highest eigenvalues
                        # are usually incorrect
                        # ask Geoffroy Hautier for more details
                        nb_bands = int(math.floor(self._bs.nb_bands*(1-self.cb_cut)))
                        for j in range(nb_bands):
                            eigs.append(
                                Energy(self._bs.bands[Spin(spin)][j][i] -
                                       self._bs.efermi, "eV").to("Ry"))
                    eigs.sort()

                    if self.run_type == "DOS" and self._bs.is_spin_polarized:
                        eigs.insert(0, self._ll)
                        eigs.append(self._hl)

                    f.write("%12.8f %12.8f %12.8f %d\n"
                            % (kpt.frac_coords[0],
                               kpt.frac_coords[1],
                               kpt.frac_coords[2],
                               len(eigs)))

                    for j in range(len(eigs)):
                        f.write("%18.8f\n" % (float(eigs[j])))

    def write_struct(self, output_file):
        if self._symprec != None:
            sym = SpacegroupAnalyzer(self._bs.structure, symprec=self._symprec)
        elif self._symprec == None:
            pass
        
        with open(output_file, 'w') as f:
            if self._symprec != None:
                f.write("{} {}\n".format(self._bs.structure.composition.formula,
                                        sym.get_space_group_symbol()))
            elif self._symprec == None:
                f.write("{} {}\n".format(self._bs.structure.composition.formula,
                                        "symmetries disabled"))

            f.write("{}\n".format("\n".join(
                [" ".join(["%.5f" % Length(i, "ang").to("bohr") for i in row])
                 for row in self._bs.structure.lattice.matrix])))

            if self._symprec != None:
                ops = sym.get_symmetry_dataset()['rotations'] 
            elif self._symprec == None:
                ops = [[[1,0,0],[0,1,0],[0,0,1]]]
            f.write("{}\n".format(len(ops)))

            for c in ops:
                for row in c:
                    f.write("{}\n".format(" ".join(str(i) for i in row)))

    def write_def(self, output_file):
        # This function is useless in std version of BoltzTraP code
        # because x_trans script overwrite BoltzTraP.def
        with open(output_file, 'w') as f:
            so = ""
            if self._bs.is_spin_polarized or self.soc:
                so = "so"
            f.write("5, 'boltztrap.intrans',      'old',    'formatted',0\n" +
                    "6,'boltztrap.outputtrans',      'unknown',    "
                    "'formatted',0\n" +
                    "20,'boltztrap.struct',         'old',    'formatted',0\n"
                    + "10,'boltztrap.energy" + so + "',         'old',    "
                                                    "'formatted',0\n" +
                    "48,'boltztrap.engre',         'unknown',    "
                    "'unformatted',0\n" +
                    "49,'boltztrap.transdos',        'unknown',    "
                    "'formatted',0\n" +
                    "50,'boltztrap.sigxx',        'unknown',    'formatted',"
                    "0\n" +
                    "51,'boltztrap.sigxxx',        'unknown',    'formatted',"
                    "0\n" +
                    "21,'boltztrap.trace',           'unknown',    "
                    "'formatted',0\n" +
                    "22,'boltztrap.condtens',           'unknown',    "
                    "'formatted',0\n" +
                    "24,'boltztrap.halltens',           'unknown',    "
                    "'formatted',0\n" +
                    "30,'boltztrap_BZ.cube',           'unknown',    "
                    "'formatted',0\n")

    def write_proj(self, output_file_proj, output_file_def):
        # This function is useless in std version of BoltzTraP code
        # because x_trans script overwrite BoltzTraP.def
        for oi, o in enumerate(Orbital):
            for site_nb in range(0, len(self._bs.structure.sites)):
                if oi < len(self._bs.projections[Spin.up][0][0]):
                    with open(output_file_proj + "_" + str(site_nb) + "_" + str(
                            o),
                              'w') as f:
                        f.write(self._bs.structure.composition.formula + "\n")
                        f.write(str(len(self._bs.kpoints)) + "\n")
                        for i in range(len(self._bs.kpoints)):
                            tmp_proj = []
                            for j in range(
                                    int(math.floor(self._bs.nb_bands*(1-self.cb_cut)))):
                                tmp_proj.append(
                                    self._bs.projections[Spin(self.spin)][j][
                                        i][oi][site_nb])
                            # TODO deal with the sorting going on at
                            # the energy level!!!
                            # tmp_proj.sort()

                            if self.run_type == "DOS" and \
                                    self._bs.is_spin_polarized:
                                tmp_proj.insert(0, self._ll)
                                tmp_proj.append(self._hl)

                            f.write("%12.8f %12.8f %12.8f %d\n"
                                    % (self._bs.kpoints[i].frac_coords[0],
                                       self._bs.kpoints[i].frac_coords[1],
                                       self._bs.kpoints[i].frac_coords[2],
                                       len(tmp_proj)))
                            for j in range(len(tmp_proj)):
                                f.write("%18.8f\n" % float(tmp_proj[j]))
        with open(output_file_def, 'w') as f:
            so = ""
            if self._bs.is_spin_polarized:
                so = "so"
            f.write("5, 'boltztrap.intrans',      'old',    'formatted',0\n" +
                    "6,'boltztrap.outputtrans',      'unknown',    "
                    "'formatted',0\n" +
                    "20,'boltztrap.struct',         'old',    'formatted',0\n"
                    + "10,'boltztrap.energy" + so + "',         'old',    "
                                                    "'formatted',0\n" +
                    "48,'boltztrap.engre',         'unknown',    "
                    "'unformatted',0\n" +
                    "49,'boltztrap.transdos',        'unknown',    "
                    "'formatted',0\n" +
                    "50,'boltztrap.sigxx',        'unknown',    'formatted',"
                    "0\n" +
                    "51,'boltztrap.sigxxx',        'unknown',    'formatted',"
                    "0\n" +
                    "21,'boltztrap.trace',           'unknown',    "
                    "'formatted',0\n" +
                    "22,'boltztrap.condtens',           'unknown',    "
                    "'formatted',0\n" +
                    "24,'boltztrap.halltens',           'unknown',    "
                    "'formatted',0\n" +
                    "30,'boltztrap_BZ.cube',           'unknown',    "
                    "'formatted',0\n")
            i = 1000
            for oi, o in enumerate(Orbital):
                for site_nb in range(0, len(self._bs.structure.sites)):
                    if oi < len(self._bs.projections[Spin.up][0][0]):
                        f.write(str(i) + ",\'" + "boltztrap.proj_" + str(
                            site_nb) + "_" + str(o.name) +
                                "\' \'old\', \'formatted\',0\n")
                        i += 1

    def write_intrans(self, output_file):
        setgap = 1 if self.scissor > 0.0001 else 0

        if self.run_type == "BOLTZ" or self.run_type == "DOS":
            with open(output_file, 'w') as fout:
                fout.write("GENE          # use generic interface\n")
                fout.write(
                    "1 0 %d %f         # iskip (not presently used) idebug "
                    "setgap shiftgap \n"
                    % (setgap, Energy(self.scissor, "eV").to("Ry")))
                fout.write(
                    "0.0 %f %f %6.1f     # Fermilevel (Ry),energygrid,energy "
                    "span around Fermilevel, number of electrons\n"
                    % (Energy(self.energy_grid, "eV").to("Ry"),
                       Energy(self.energy_span_around_fermi, "eV").to("Ry"),
                       self._nelec))
                fout.write(
                    "CALC                    # CALC (calculate expansion "
                    "coeff), NOCALC read from file\n")
                fout.write(
                    "%d                        # lpfac, number of latt-points "
                    "per k-point\n" % self.lpfac)
                fout.write(
                    "%s                     # run mode (only BOLTZ is "
                    "supported)\n" % self.run_type)
                fout.write(
                    ".15                       # (efcut) energy range of "
                    "chemical potential\n")
                fout.write(
                    "{} {}                  # Tmax, temperature grid\n". \
                        format(self.tmax, self.tgrid))
                fout.write(
                    "-1.  # energyrange of bands given DOS output sig_xxx and "
                    "dos_xxx (xxx is band number)\n")
                fout.write(self.dos_type + "\n")  # e.g., HISTO or TETRA
                fout.write("{} {} {} 0 0 0\n".format(
                    self.tauref, self.tauexp, self.tauen))
                fout.write("{}\n".format(2 * len(self.doping)))

                for d in self.doping:
                    fout.write(str(d) + "\n")
                for d in self.doping:
                    fout.write(str(-d) + "\n")

        elif self.run_type == "FERMI":
            with open(output_file, 'w') as fout:
                fout.write("GENE          # use generic interface\n")
                fout.write(
                    "1 0 0 0.0         # iskip (not presently used) idebug "
                    "setgap shiftgap \n")
                fout.write(
                    "0.0 %f 0.1 %6.1f     # Fermilevel (Ry),energygrid,"
                    "energy span around Fermilevel, "
                    "number of electrons\n"
                    % (Energy(self.energy_grid, "eV").to("Ry"), self._nelec))
                fout.write(
                    "CALC                    # CALC (calculate expansion "
                    "coeff), NOCALC read from file\n")
                fout.write(
                    "%d                        # lpfac, number of latt-points "
                    "per k-point\n" % self.lpfac)
                fout.write(
                    "FERMI                     # run mode (only BOLTZ is "
                    "supported)\n")
                fout.write(str(1) +
                           "                        # actual band selected: " +
                           str(self.band_nb + 1) + " spin: " + str(self.spin))

        elif self.run_type == "BANDS":
            if self.kpt_line is None:
                kpath = HighSymmKpath(self._bs.structure)
                self.kpt_line = [Kpoint(k, self._bs.structure.lattice) for k
                                 in
                                 kpath.get_kpoints(coords_are_cartesian=False)[
                                     0]]
                self.kpt_line = [kp.frac_coords for kp in self.kpt_line]
            elif type(self.kpt_line[0]) == Kpoint:
                self.kpt_line = [kp.frac_coords for kp in self.kpt_line]

            with open(output_file, 'w') as fout:
                fout.write("GENE          # use generic interface\n")
                fout.write(
                    "1 0 %d %f         # iskip (not presently used) idebug "
                    "setgap shiftgap \n"
                    % (setgap, Energy(self.scissor, "eV").to("Ry")))
                fout.write(
                    "0.0 %f %f %6.1f     # Fermilevel (Ry),energygrid,energy "
                    "span around Fermilevel, "
                    "number of electrons\n"
                    % (Energy(self.energy_grid, "eV").to("Ry"),
                       Energy(self.energy_span_around_fermi, "eV").to("Ry"),
                       self._nelec))
                fout.write(
                    "CALC                    # CALC (calculate expansion "
                    "coeff), NOCALC read from file\n")
                fout.write(
                    "%d                        # lpfac, number of latt-points "
                    "per k-point\n" % self.lpfac)
                fout.write(
                    "BANDS                     # run mode (only BOLTZ is "
                    "supported)\n")
                fout.write("P " + str(len(self.kpt_line)) + "\n")
                for kp in self.kpt_line:
                    fout.writelines([str(k) + " " for k in kp])
                    fout.write('\n')

    def write_input(self, output_dir):
        if self._bs.is_spin_polarized or self.soc:
            self.write_energy(os.path.join(output_dir, "boltztrap.energyso"))
        else:
            self.write_energy(os.path.join(output_dir, "boltztrap.energy"))

        self.write_struct(os.path.join(output_dir, "boltztrap.struct"))
        self.write_intrans(os.path.join(output_dir, "boltztrap.intrans"))
        self.write_def(os.path.join(output_dir, "BoltzTraP.def"))

        if len(self.bs.projections) != 0 and self.run_type == "DOS":
            self.write_proj(os.path.join(output_dir, "boltztrap.proj"),
                            os.path.join(output_dir, "BoltzTraP.def"))

    def run(self, path_dir=None, convergence=True, write_input=True,
            clear_dir=False, max_lpfac=150, min_egrid=0.00005):
        """
        Write inputs (optional), run BoltzTraP, and ensure
        convergence (optional)
        Args:
            path_dir (str): directory in which to run BoltzTraP
            convergence (bool): whether to check convergence and make
                corrections if needed
            write_input: (bool) whether to write input files before the run
                (required for convergence mode)
            clear_dir: (bool) whether to remove all files in the path_dir
                before starting
            max_lpfac: (float) maximum lpfac value to try before reducing egrid
                in convergence mode
            min_egrid: (float) minimum egrid value to try before giving up in
                convergence mode

        Returns:

        """

        # TODO: consider making this a part of custodian rather than pymatgen
        # A lot of this functionality (scratch dirs, handlers, monitors)
        # is built into custodian framework

        if convergence and not write_input:
            raise ValueError("Convergence mode requires write_input to be "
                             "true")

        if self.run_type in ("BANDS", "DOS", "FERMI"):
            convergence = False
            if self.lpfac > max_lpfac:
                max_lpfac = self.lpfac

        if self.run_type == "BANDS" and self.bs.is_spin_polarized:
            print("Reminder: for run_type " + str(
                self.run_type) + ", spin component are not separated! "
                                 "(you have a spin polarized band structure)")

        if self.run_type in ("FERMI", "DOS") and self.spin is None:
            if self.bs.is_spin_polarized:
                raise BoltztrapError(
                    "Spin parameter must be specified for spin polarized "
                    "band structures!")
            else:
                self.spin = 1

        dir_bz_name = "boltztrap"
        if path_dir is None:
            temp_dir = tempfile.mkdtemp()
            path_dir = os.path.join(temp_dir, dir_bz_name)
        else:
            path_dir = os.path.abspath(
                os.path.join(path_dir, dir_bz_name))

        if not os.path.exists(path_dir):
            os.mkdir(path_dir)
        elif clear_dir:
            for c in os.listdir(path_dir):
                os.remove(os.path.join(path_dir, c))

        FORMAT = "%(message)s"
        logging.basicConfig(level=logging.INFO, format=FORMAT,
                            filename=os.path.join(path_dir, "../boltztrap.out"))

        with cd(path_dir):
            lpfac_start = self.lpfac
            converged = False

            while self.energy_grid >= min_egrid and not converged:
                self.lpfac = lpfac_start
                if time.time() - self.start_time > self.timeout:
                    raise BoltztrapError("no doping convergence after timeout "
                                         "of {} s".format(self.timeout))

                logging.info("lpfac, energy_grid: {} {}".format(self.lpfac, self.energy_grid))

                while self.lpfac <= max_lpfac and not converged:
                    if time.time() - self.start_time > self.timeout:
                        raise BoltztrapError("no doping convergence after "
                                        "timeout of {} s".format(self.timeout))

                    if write_input:
                        self.write_input(path_dir)

                    bt_exe = ["x_trans", "BoltzTraP"]
                    if self._bs.is_spin_polarized or self.soc:
                        bt_exe.append("-so")

                    p = subprocess.Popen(bt_exe, stdout=subprocess.PIPE,
                                         stdin=subprocess.PIPE,
                                         stderr=subprocess.PIPE)
                    p.wait()

                    for c in p.communicate():
                        logging.info(c.decode())
                        if "error in factorization" in c.decode():
                            raise BoltztrapError("error in factorization")

                    warning = ""

                    with open(os.path.join(path_dir,
                                           dir_bz_name + ".outputtrans")) as f:
                        for l in f:
                            if "Option unknown" in l:
                                raise BoltztrapError(
                                    "DOS mode needs a custom version of "
                                    "BoltzTraP code is needed")
                            if "WARNING" in l:
                                warning = l
                                break
                            if "Error - Fermi level was not found" in l:
                                warning = l
                                break

                    if not warning and convergence:
                        # check convergence for warning
                        analyzer = BoltztrapAnalyzer.from_files(path_dir)
                        for doping in ['n', 'p']:
                            for c in analyzer.mu_doping[doping]:
                                if len(analyzer.mu_doping[doping][c]) != len(
                                        analyzer.doping[doping]):
                                    warning = "length of mu_doping array is " \
                                              "incorrect"
                                    break

                                if doping == 'p' and \
                                                sorted(
                                                    analyzer.mu_doping[doping][
                                                        c], reverse=True) != \
                                                analyzer.mu_doping[doping][c]:
                                    warning = "sorting of mu_doping array " \
                                              "incorrect for p-type"
                                    break

                                # ensure n-type doping sorted correctly
                                if doping == 'n' and sorted(
                                        analyzer.mu_doping[doping][c]) != \
                                        analyzer.mu_doping[doping][c]:
                                    warning = "sorting of mu_doping array " \
                                              "incorrect for n-type"
                                    break

                    if warning:
                        self.lpfac += 10
                        logging.warn("Warning detected: {}! Increase lpfac to "
                                     "{}".format(warning, self.lpfac))

                    else:
                        converged = True

                if not converged:
                    self.energy_grid /= 10
                    logging.info("Could not converge with max lpfac; "
                                 "Decrease egrid to {}".format(self.energy_grid))

            if not converged:
                raise BoltztrapError(
                    "Doping convergence not reached with lpfac=" + str(
                        self.lpfac) + ", energy_grid=" + str(self.energy_grid))

            return path_dir

    def as_dict(self):
        results =  {"@module": self.__class__.__module__,
                "@class": self.__class__.__name__,
                "lpfac": self.lpfac,
                "bs": self.bs.as_dict(),
                "nelec": self._nelec,
                "dos_type": self.dos_type,
                "run_type": self.run_type,
                "band_nb": self.band_nb,
                "spin": self.spin,
                "cond_band": self.cond_band,
                "tauref": self.tauref,
                "tauexp": self.tauexp,
                "tauen": self.tauen,
                "soc": self.soc,
                "kpt_line": self.kpt_line,
                "doping": self.doping,
                "energy_span_around_fermi": self.energy_span_around_fermi,
                "scissor": self.scissor,
                "tmax": self.tmax,
                "tgrid": self.tgrid,
                "symprec": self._symprec
                }
        return jsanitize(results)


class BoltztrapError(Exception):
    """
    Exception class for boltztrap.
    Raised when the boltztrap gives an error
    """

    def __init__(self, msg):
        self.msg = msg
        logging.error(self.msg)

    def __str__(self):
        return "BoltztrapError : " + self.msg


class BoltztrapAnalyzer:
    """
    Class used to store all the data from a boltztrap run
    """

    def __init__(self, gap=None, mu_steps=None, cond=None, seebeck=None,
                 kappa=None, hall=None, doping=None,
                 mu_doping=None, seebeck_doping=None, cond_doping=None,
                 kappa_doping=None,
                 hall_doping=None, intrans=None, dos=None, dos_partial=None,
                 carrier_conc=None, vol=None, warning=None,
                 bz_bands=None, bz_kpoints=None, fermi_surface_data=None):
        """
        Constructor taking directly all the data generated by Boltztrap. You
        won't probably use it directly but instead use the from_files and
        from_dict methods.

        Args:
            gap: The gap after interpolation in eV
            mu_steps: The steps of electron chemical potential (or Fermi
                level) in eV.
            cond: The electronic conductivity tensor divided by a constant
                relaxation time (sigma/tau) at different temperature and
                fermi levels.
                The format is {temperature: [array of 3x3 tensors at each
                fermi level in mu_steps]}. The units are 1/(Ohm*m*s).
            seebeck: The Seebeck tensor at different temperatures and fermi
                levels. The format is {temperature: [array of 3x3 tensors at
                each fermi level in mu_steps]}. The units are V/K
            kappa: The electronic thermal conductivity tensor divided by a
                constant relaxation time (kappa/tau) at different temperature
                and fermi levels. The format is {temperature: [array of 3x3
                tensors at each fermi level in mu_steps]}
                The units are W/(m*K*s)
            hall: The hall tensor at different temperature and fermi levels
                The format is {temperature: [array of 27 coefficients list at
                each fermi level in mu_steps]}
                The units are m^3/C
            doping: The different doping levels that have been given to
                Boltztrap. The format is {'p':[],'n':[]} with an array of
                doping levels. The units are cm^-3
            mu_doping: Gives the electron chemical potential (or Fermi level)
                for a given set of doping.
                Format is {'p':{temperature: [fermi levels],'n':{temperature:
                [fermi levels]}}
                the fermi level array is ordered according to the doping
                levels in doping units for doping are in cm^-3 and for Fermi
                level in eV
            seebeck_doping: The Seebeck tensor at different temperatures and
                doping levels. The format is {'p': {temperature: [Seebeck
                tensors]}, 'n':{temperature: [Seebeck tensors]}}
                The [Seebeck tensors] array is ordered according to the
                doping levels in doping units for doping are in cm^-3 and for
                Seebeck in V/K
            cond_doping: The electronic conductivity tensor divided by a
                constant relaxation time (sigma/tau) at different
                temperatures and doping levels
                The format is {'p':{temperature: [conductivity tensors]},
                'n':{temperature: [conductivity tensors]}}
                The [conductivity tensors] array is ordered according to the
                doping levels in doping units for doping are in cm^-3 and for
                conductivity in 1/(Ohm*m*s)
            kappa_doping: The thermal conductivity tensor divided by a constant
                relaxation time (kappa/tau) at different temperatures and
                doping levels.
                The format is {'p':{temperature: [thermal conductivity
                tensors]},'n':{temperature: [thermal conductivity tensors]}}
                The [thermal conductivity tensors] array is ordered according
                to the doping levels in doping units for doping are in cm^-3
                and for thermal conductivity in W/(m*K*s)
            hall_doping: The Hall tensor at different temperatures and doping
                levels.
                The format is {'p':{temperature: [Hall tensors]},
                'n':{temperature: [Hall tensors]}}
                The [Hall tensors] array is ordered according to the doping
                levels in doping and each Hall tensor is represented by a 27
                coefficients list.
                The units are m^3/C
            intrans: a dictionary of inputs e.g. {"scissor": 0.0}
            carrier_conc: The concentration of carriers in electron (or hole)
                per unit cell
            dos: The dos computed by Boltztrap given as a pymatgen Dos object
            dos_partial: Data for the partial DOS projected on sites and
                orbitals
            vol: Volume of the unit cell in angstrom cube (A^3)
            warning: string if BoltzTraP outputted a warning, else None
            bz_bands: Data for interpolated bands on a k-point line
                (run_type=BANDS)
            bz_kpoints: k-point in reciprocal coordinates for interpolated
                bands (run_type=BANDS)
            fermi_surface_data: energy values in a 3D grid imported from the
                output .cube file.
        """
        self.gap = gap
        self.mu_steps = mu_steps
        self._cond = cond
        self._seebeck = seebeck
        self._kappa = kappa
        self._hall = hall
        self.warning = warning
        self.doping = doping
        self.mu_doping = mu_doping
        self._seebeck_doping = seebeck_doping
        self._cond_doping = cond_doping
        self._kappa_doping = kappa_doping
        self._hall_doping = hall_doping
        self.intrans = intrans
        self._carrier_conc = carrier_conc
        self.dos = dos
        self.vol = vol
        self._dos_partial = dos_partial
        self._bz_bands = bz_bands
        self._bz_kpoints = bz_kpoints
        self.fermi_surface_data = fermi_surface_data

    def get_symm_bands(self, structure, efermi, kpt_line=None,
                       labels_dict=None):
        """
            Function useful to read bands from Boltztrap output and get a
            BandStructureSymmLine object comparable with that one from a DFT
            calculation (if the same kpt_line is provided). Default kpt_line
            and labels_dict is the standard path of high symmetry k-point for
            the specified structure. They could be extracted from the
            BandStructureSymmLine object that you want to compare with. efermi
            variable must be specified to create the BandStructureSymmLine
            object (usually it comes from DFT or Boltztrap calc)
        """
        try:
            if kpt_line is None:
                kpath = HighSymmKpath(structure)
                kpt_line = [Kpoint(k, structure.lattice.reciprocal_lattice) for
                            k in
                            kpath.get_kpoints(coords_are_cartesian=False)[0]]
                labels_dict = {l: k for k, l in zip(
                    *kpath.get_kpoints(coords_are_cartesian=False)) if l}
                kpt_line = [kp.frac_coords for kp in kpt_line]
            elif type(kpt_line[0]) == Kpoint:
                kpt_line = [kp.frac_coords for kp in kpt_line]
                labels_dict = {k: labels_dict[k].frac_coords for k in
                               labels_dict}

            idx_list = []
            #       kpt_dense=np.array([kp for kp in self._bz_kpoints])
            for i, kp in enumerate(kpt_line):
                w = []
                prec = 1e-05
                while len(w) == 0:
                    w = np.where(np.all(
                        np.abs(kp - self._bz_kpoints) < [prec] * 3,
                        axis=1))[0]
                    prec *= 10

                # print( prec )
                idx_list.append([i, w[0]])

                # if len(w)>0:
                #     idx_list.append([i,w[0]])
                # else:
                #     w=np.where(np.all(np.abs(kp.frac_coords-self._bz_kpoints)
                # <[1e-04,1e-04,1e-04],axis=1))[0]
                #     idx_list.append([i,w[0]])

            idx_list = np.array(idx_list)
            # print( idx_list.shape )

            bands_dict = {Spin.up: (self._bz_bands * Energy(1, "Ry").to(
                "eV") + efermi).T[:, idx_list[:, 1]].tolist()}
            # bz_kpoints = bz_kpoints[idx_list[:,1]].tolist()

            sbs = BandStructureSymmLine(kpt_line, bands_dict,
                                        structure.lattice.reciprocal_lattice,
                                        efermi,
                                        labels_dict=labels_dict)

            return sbs

        except:
            raise BoltztrapError(
                "Bands are not in output of BoltzTraP.\nBolztrapRunner must "
                "be run with run_type=BANDS")

    @staticmethod
    def check_acc_bzt_bands(sbs_bz, sbs_ref, warn_thr=(0.03, 0.03)):
        """
            Compare sbs_bz BandStructureSymmLine calculated with boltztrap with
            the sbs_ref BandStructureSymmLine as reference (from MP for
            instance), computing correlation and energy difference for eight bands
            around the gap (semiconductors) or fermi level (metals).
            warn_thr is a threshold to get a warning in the accuracy of Boltztap
            interpolated bands.
            Return a dictionary with these keys:
            - "N": the index of the band compared; inside each there are:
                - "Corr": correlation coefficient for the 8 compared bands
                - "Dist": energy distance for the 8 compared bands
                - "branch_name": energy distance for that branch
            - "avg_corr": average of correlation coefficient over the 8 bands
            - "avg_dist": average of energy distance over the 8 bands
            - "nb_list": list of indexes of the 8 compared bands
            - "acc_thr": list of two float corresponing to the two warning
                         thresholds in input
            - "acc_err": list of two bools:
                         True if the avg_corr > warn_thr[0], and
                         True if the avg_dist > warn_thr[1]
            See also compare_sym_bands function doc
        """
        if not sbs_ref.is_metal() and not sbs_bz.is_metal():
            vbm_idx = sbs_bz.get_vbm()['band_index'][Spin.up][-1]
            cbm_idx = sbs_bz.get_cbm()['band_index'][Spin.up][0]
            nb_list = range(vbm_idx - 3, cbm_idx + 4)

        else:
            bnd_around_efermi = []
            delta = 0
            spin = list(sbs_bz.bands.keys())[0]
            while len(bnd_around_efermi) < 8 and delta < 100:
                delta += 0.1
                bnd_around_efermi = []
                for nb in range(len(sbs_bz.bands[spin])):
                    for kp in range(len(sbs_bz.bands[spin][nb])):
                        if abs(sbs_bz.bands[spin][nb][
                                   kp] - sbs_bz.efermi) < delta:
                            bnd_around_efermi.append(nb)
                            break
            if len(bnd_around_efermi) < 8:
                print("Warning! check performed on " + str(
                    len(bnd_around_efermi)))
                nb_list = bnd_around_efermi
            else:
                nb_list = bnd_around_efermi[:8]

        # print(nb_list)
        bcheck = compare_sym_bands(sbs_bz, sbs_ref, nb_list)
        # print(bcheck)
        acc_err = [False, False]
        avg_corr = sum([item[1]['Corr'] for item in bcheck.iteritems()]) / 8
        avg_distance = sum([item[1]['Dist'] for item in bcheck.iteritems()]) / 8

        if avg_corr > warn_thr[0]: acc_err[0] = True
        if avg_distance > warn_thr[0]: acc_err[1] = True

        bcheck['avg_corr'] = avg_corr
        bcheck['avg_distance'] = avg_distance
        bcheck['acc_err'] = acc_err
        bcheck['acc_thr'] = warn_thr
        bcheck['nb_list'] = nb_list

        if True in acc_err:
            print("Warning! some bands around gap are not accurate")

        return bcheck

    def get_seebeck(self, output='eigs', doping_levels=True):
        """
            Gives the seebeck coefficient (microV/K) in either a
            full 3x3 tensor form, as 3 eigenvalues, or as the average value
            (trace/3.0) If doping_levels=True, the results are given at
            different p and n doping
            levels (given by self.doping), otherwise it is given as a series
            of electron chemical potential values

            Args:
                output (string): the type of output. 'tensor' give the full
                3x3 tensor, 'eigs' its 3 eigenvalues and
                'average' the average of the three eigenvalues
                doping_levels (boolean): True for the results to be given at
                different doping levels, False for results
                at different electron chemical potentials

            Returns:
                If doping_levels=True, a dictionary {temp:{'p':[],'n':[]}}.
                The 'p' links to Seebeck at p-type doping
                and 'n' to the Seebeck at n-type doping. Otherwise, returns a
                {temp:[]} dictionary
                The result contains either the sorted three eigenvalues of
                the symmetric
                Seebeck tensor (output='eigs') or a full tensor (3x3 array) (
                output='tensor') or as an average
                (output='average').

                units are microV/K
        """
        return BoltztrapAnalyzer._format_to_output(self._seebeck,
                                                   self._seebeck_doping,
                                                   output,
                                                   doping_levels, 1e6)

    def get_conductivity(self, output='eigs', doping_levels=True,
                         relaxation_time=1e-14):
        """
            Gives the conductivity (1/Ohm*m) in either a full 3x3 tensor
            form, as 3 eigenvalues, or as the average value
            (trace/3.0) If doping_levels=True, the results are given at
            different p and n doping
            levels (given by self.doping), otherwise it is given as a series
            of electron chemical potential values

            Args:
                output (string): the type of output. 'tensor' give the full
                3x3 tensor, 'eigs' its 3 eigenvalues and
                'average' the average of the three eigenvalues
                doping_levels (boolean): True for the results to be given at
                different doping levels, False for results
                at different electron chemical potentials
                relaxation_time (float): constant relaxation time in secs

            Returns:
                If doping_levels=True, a dictionary {temp:{'p':[],'n':[]}}.
                The 'p' links to conductivity
                at p-type doping and 'n' to the conductivity at n-type
                doping. Otherwise,
                returns a {temp:[]} dictionary. The result contains either
                the sorted three eigenvalues of the symmetric
                conductivity tensor (format='eigs') or a full tensor (3x3
                array) (output='tensor') or as an average
                (output='average').
                The result includes a given constant relaxation time

                units are 1/Ohm*m
        """
        return BoltztrapAnalyzer._format_to_output(self._cond,
                                                   self._cond_doping, output,
                                                   doping_levels,
                                                   relaxation_time)

    def get_power_factor(self, output='eigs', doping_levels=True,
                         relaxation_time=1e-14):
        """
        Gives the power factor (Seebeck^2 * conductivity) in units
        microW/(m*K^2) in either a full 3x3 tensor form,
        as 3 eigenvalues, or as the average value (trace/3.0) If
        doping_levels=True, the results are given at
        different p and n doping levels (given by self.doping), otherwise it
        is given as a series of
        electron chemical potential values

        Args:
            output (string): the type of output. 'tensor' give the full 3x3
            tensor, 'eigs' its 3 eigenvalues and
            'average' the average of the three eigenvalues
            doping_levels (boolean): True for the results to be given at
            different doping levels, False for results
            at different electron chemical potentials
            relaxation_time (float): constant relaxation time in secs

        Returns:
            If doping_levels=True, a dictionnary {temp:{'p':[],'n':[]}}. The
            'p' links to power factor
            at p-type doping and 'n' to the conductivity at n-type doping.
            Otherwise,
            returns a {temp:[]} dictionary. The result contains either the
            sorted three eigenvalues of the symmetric
            power factor tensor (format='eigs') or a full tensor (3x3 array) (
            output='tensor') or as an average
            (output='average').
            The result includes a given constant relaxation time

            units are microW/(m K^2)
        """
        result = None
        result_doping = None
        if doping_levels:
            result_doping = {doping: {t: [] for t in
                                      self._seebeck_doping[doping]} for
                             doping in self._seebeck_doping}

            for doping in result_doping:
                for t in result_doping[doping]:
                    for i in range(len(self.doping[doping])):
                        full_tensor = np.dot(self._cond_doping[doping][t][i],
                                             np.dot(
                                                 self._seebeck_doping[doping][
                                                     t][i],
                                                 self._seebeck_doping[doping][
                                                     t][i]))
                        result_doping[doping][t].append(full_tensor)

        else:
            result = {t: [] for t in self._seebeck}
            for t in result:
                for i in range(len(self.mu_steps)):
                    full_tensor = np.dot(self._cond[t][i],
                                         np.dot(self._seebeck[t][i],
                                                self._seebeck[t][i]))
                    result[t].append(full_tensor)

        return BoltztrapAnalyzer._format_to_output(result, result_doping,
                                                   output, doping_levels,
                                                   multi=1e6 * relaxation_time)

    def get_thermal_conductivity(self, output='eigs', doping_levels=True,
                                 k_el=True, relaxation_time=1e-14):
        """
        Gives the electronic part of the thermal conductivity in either a
        full 3x3 tensor form,
        as 3 eigenvalues, or as the average value (trace/3.0) If
        doping_levels=True, the results are given at
        different p and n doping levels (given by self.doping), otherwise it
        is given as a series of
        electron chemical potential values

        Args:
            output (string): the type of output. 'tensor' give the full 3x3
            tensor, 'eigs' its 3 eigenvalues and
            'average' the average of the three eigenvalues
            doping_levels (boolean): True for the results to be given at
            different doping levels, False for results
            at different electron chemical potentials
            k_el (boolean): True for k_0-PF*T, False for k_0
            relaxation_time (float): constant relaxation time in secs

        Returns:
            If doping_levels=True, a dictionary {temp:{'p':[],'n':[]}}. The
            'p' links to thermal conductivity
            at p-type doping and 'n' to the thermal conductivity at n-type
            doping. Otherwise,
            returns a {temp:[]} dictionary. The result contains either the
            sorted three eigenvalues of the symmetric
            conductivity tensor (format='eigs') or a full tensor (3x3 array) (
            output='tensor') or as an average
            (output='average').
            The result includes a given constant relaxation time

            units are W/mK
        """
        result = None
        result_doping = None
        if doping_levels:
            result_doping = {doping: {t: [] for t in
                                      self._seebeck_doping[doping]} for
                             doping in self._seebeck_doping}
            for doping in result_doping:
                for t in result_doping[doping]:
                    for i in range(len(self.doping[doping])):
                        if k_el:
                            pf_tensor = np.dot(self._cond_doping[doping][t][i],
                                               np.dot(
                                                   self._seebeck_doping[doping][
                                                       t][i],
                                                   self._seebeck_doping[doping][
                                                       t][i]))
                            result_doping[doping][t].append((
                                self._kappa_doping[doping][t][
                                    i] - pf_tensor * t))
                        else:
                            result_doping[doping][t].append((
                                self._kappa_doping[doping][t][i]))
        else:
            result = {t: [] for t in self._seebeck}
            for t in result:
                for i in range(len(self.mu_steps)):
                    if k_el:
                        pf_tensor = np.dot(self._cond[t][i],
                                           np.dot(self._seebeck[t][i],
                                                  self._seebeck[t][i]))
                        result[t].append((self._kappa[t][i] - pf_tensor * t))
                    else:
                        result[t].append((self._kappa[t][i]))

        return BoltztrapAnalyzer._format_to_output(result, result_doping,
                                                   output, doping_levels,
                                                   multi=relaxation_time)

    def get_zt(self, output='eigs', doping_levels=True, relaxation_time=1e-14,
               kl=1.0):
        """
        Gives the ZT coefficient (S^2*cond*T/thermal cond) in either a full
        3x3 tensor form,
        as 3 eigenvalues, or as the average value (trace/3.0) If
        doping_levels=True, the results are given at
        different p and n doping levels (given by self.doping), otherwise it
        is given as a series of
        electron chemical potential values. We assume a constant relaxation
        time and a constant
        lattice thermal conductivity

        Args:
            output (string): the type of output. 'tensor' give the full 3x3
            tensor, 'eigs' its 3 eigenvalues and
            'average' the average of the three eigenvalues
            doping_levels (boolean): True for the results to be given at
            different doping levels, False for results
            at different electron chemical potentials
            relaxation_time (float): constant relaxation time in secs
            k_l (float): lattice thermal cond in W/(m*K)

        Returns:
            If doping_levels=True, a dictionary {temp:{'p':[],'n':[]}}. The
            'p' links to ZT
            at p-type doping and 'n' to the ZT at n-type doping. Otherwise,
            returns a {temp:[]} dictionary. The result contains either the
            sorted three eigenvalues of the symmetric
            ZT tensor (format='eigs') or a full tensor (3x3 array) (
            output='tensor') or as an average
            (output='average').
            The result includes a given constant relaxation time and lattice
            thermal conductivity
        """
        result = None
        result_doping = None
        if doping_levels:
            result_doping = {doping: {t: [] for t in
                                      self._seebeck_doping[doping]} for
                             doping in self._seebeck_doping}

            for doping in result_doping:
                for t in result_doping[doping]:
                    for i in range(len(self.doping[doping])):
                        pf_tensor = np.dot(self._cond_doping[doping][t][i],
                                           np.dot(
                                               self._seebeck_doping[doping][t][
                                                   i],
                                               self._seebeck_doping[doping][t][
                                                   i]))
                        thermal_conduct = (self._kappa_doping[doping][t][i]
                                           - pf_tensor * t) * relaxation_time
                        result_doping[doping][t].append(
                            np.dot(pf_tensor * relaxation_time * t,
                                   np.linalg.inv(
                                       thermal_conduct + kl * np.eye(3, 3))))
        else:
            result = {t: [] for t in self._seebeck}
            for t in result:
                for i in range(len(self.mu_steps)):
                    pf_tensor = np.dot(self._cond[t][i],
                                       np.dot(self._seebeck[t][i],
                                              self._seebeck[t][i]))
                    thermal_conduct = (self._kappa[t][i]
                                       - pf_tensor * t) * relaxation_time
                    result[t].append(np.dot(pf_tensor * relaxation_time * t,
                                            np.linalg.inv(
                                                thermal_conduct + kl *
                                                np.eye(3, 3))))

        return BoltztrapAnalyzer._format_to_output(result, result_doping,
                                                   output, doping_levels)

    def get_average_eff_mass(self, output='eigs', doping_levels=True):
        """
        Gives the average effective mass tensor. We call it average because
        it takes into account all the bands
        and regions in the Brillouin zone. This is different than the standard
        textbook effective mass which relates
        often to only one (parabolic) band.
        The average effective mass tensor is defined as the integrated
        average of the second derivative of E(k)
        This effective mass tensor takes into account:
        -non-parabolicity
        -multiple extrema
        -multiple bands

        For more information about it. See:

        Hautier, G., Miglio, A., Waroquiers, D., Rignanese, G., & Gonze,
        X. (2014).
        How Does Chemistry Influence Electron Effective Mass in Oxides?
        A High-Throughput Computational Analysis. Chemistry of Materials,
        26(19), 5447–5458. doi:10.1021/cm404079a

        or

        Hautier, G., Miglio, A., Ceder, G., Rignanese, G.-M., & Gonze,
        X. (2013).
        Identification and design principles of low hole effective mass
        p-type transparent conducting oxides.
        Nature Communications, 4, 2292. doi:10.1038/ncomms3292

        Depending on the value of output, we have either the full 3x3
        effective mass tensor,
        its 3 eigenvalues or an average

        Args:
            output (string): 'eigs' for eigenvalues, 'tensor' for the full
            tensor and 'average' for an average (trace/3)
            doping_levels (boolean): True for the results to be given at
            different doping levels, False for results
            at different electron chemical potentials
        Returns:
            If doping_levels=True,a dictionary {'p':{temp:[]},'n':{temp:[]}}
            with an array of effective mass tensor, eigenvalues of average
            value (depending on output) for each temperature and for each
            doping level.
            The 'p' links to hole effective mass tensor and 'n' to electron
            effective mass tensor.
        """
        result = None
        result_doping = None
        conc = self.get_carrier_concentration()
        if doping_levels:
            result_doping = {doping: {t: [] for t in self._cond_doping[doping]}
                             for
                             doping in self.doping}
            for doping in result_doping:
                for temp in result_doping[doping]:
                    for i in range(len(self.doping[doping])):
                        try:
                            result_doping[doping][temp].append(np.linalg.inv(
                                np.array(self._cond_doping[doping][temp][i])) * \
                                                               self.doping[doping][
                                                                   i] * 10 ** 6 * 
                                                                    constants.e ** 2 / 
                                                                    constants.m_e)
                        except np.linalg.LinAlgError:
                            pass
        else:
            result = {t: [] for t in self._seebeck}
            for temp in result:
                for i in range(len(self.mu_steps)):
                    try:
                        cond_inv = np.linalg.inv(np.array(self._cond[temp][i]))
                    except np.linalg.LinAlgError:
                        pass
                    result[temp].append(cond_inv * \
                                        conc[temp][i] * 10 ** 6 * 
                                                constants.e ** 2 / 
                                                constants.m_e)

        return BoltztrapAnalyzer._format_to_output(result, result_doping,
                                                   output, doping_levels)
    
    def get_seebeck_eff_mass(self, output='average', temp=300, doping_levels=False,
                                Lambda=0.5):
        """
        Seebeck effective mass calculated as explained in Ref.
        Gibbs, Z. M. et al., Effective mass and fermi surface complexity factor 
        from ab initio band structure calculations. 
        npj Computational Materials 3, 8 (2017).

        Args:
            output: 'average' returns the seebeck effective mass calculated using 
                    the average of the three diagonal components of the seebeck tensor.
                    'tensor' returns the seebeck effective mass respect to the three 
                    diagonal components of the seebeck tensor.
            doping_levels: False means that the seebeck effective mass is calculated
                           for every value of the chemical potential
                           True means that the seebeck effective mass is calculated
                           for every value of the doping levels for both n and p types
            temp:   temperature of calculated seebeck.
            Lambda: fitting parameter used to model the scattering (0.5 means constant 
                    relaxation time).
        Returns:
            a list of values for the seebeck effective mass w.r.t the chemical potential,
            if doping_levels is set at False;
            a dict with n an p keys that contain a list of values for the seebeck effective
            mass w.r.t the doping levels, if doping_levels is set at True;
            if 'tensor' is selected, each element of the lists is a list containing 
            the three components of the seebeck effective mass.
        """
                
        if doping_levels:
            sbk_mass = {}
            for dt in ('n','p'):
                conc = self.doping[dt]
                seebeck = self.get_seebeck(output=output, doping_levels=True)[dt][temp]
                sbk_mass[dt] = []
                for i in range(len(conc)):
                    if output == 'average':
                        sbk_mass[dt].append(
                            seebeck_eff_mass_from_seebeck_carr(abs(seebeck[i]), 
                                                                conc[i], temp, Lambda))
                    elif output == 'tensor':
                        sbk_mass[dt].append([])
                        for j in range(3):
                            sbk_mass[dt][-1].append(
                                seebeck_eff_mass_from_seebeck_carr(abs(seebeck[i][j][j]),
                                                                    conc[i], temp, Lambda))
                
        else:
            seebeck = self.get_seebeck(output=output, doping_levels=False)[temp]
            conc = self.get_carrier_concentration()[temp]
            sbk_mass = []
            for i in range(len(conc)):
                if output == 'average':
                    sbk_mass.append(
                        seebeck_eff_mass_from_seebeck_carr(abs(seebeck[i]), 
                                                            conc[i], temp, Lambda))
                elif output == 'tensor':
                    sbk_mass.append([])
                    for j in range(3):
                        sbk_mass[-1].append(
                            seebeck_eff_mass_from_seebeck_carr(abs(seebeck[i][j][j]),
                                                                conc[i], temp, Lambda))
        return sbk_mass
    
    def get_complexity_factor(self, output='average', temp=300, doping_levels=False,
                              Lambda=0.5):
        """
        Fermi surface complexity factor respect to calculated as explained in Ref.
        Gibbs, Z. M. et al., Effective mass and fermi surface complexity factor 
        from ab initio band structure calculations. 
        npj Computational Materials 3, 8 (2017).
        
        Args:
            output: 'average' returns the complexity factor calculated using the average
                    of the three diagonal components of the seebeck and conductivity tensors.
                    'tensor' returns the complexity factor respect to the three
                    diagonal components of seebeck and conductivity tensors.
            doping_levels: False means that the complexity factor is calculated
                           for every value of the chemical potential
                           True means that the complexity factor is calculated
                           for every value of the doping levels for both n and p types
            temp:   temperature of calculated seebeck and conductivity.
            Lambda: fitting parameter used to model the scattering (0.5 means constant 
                    relaxation time).
        Returns:
            a list of values for the complexity factor w.r.t the chemical potential,
            if doping_levels is set at False;
            a dict with n an p keys that contain a list of values for the complexity factor
            w.r.t the doping levels, if doping_levels is set at True;
            if 'tensor' is selected, each element of the lists is a list containing 
            the three components of the complexity factor.
        """
        
        if doping_levels:
            cmplx_fact = {}
            for dt in ('n','p'):
                sbk_mass = self.get_seebeck_eff_mass(output, temp, True, Lambda)[dt]
                cond_mass = self.get_average_eff_mass(output=output, doping_levels=True)[dt][temp]
                
                if output == 'average':
                    cmplx_fact[dt] = [ (m_s/abs(m_c))**1.5 for m_s,m_c in zip(sbk_mass,cond_mass)]
                elif output == 'tensor':
                    cmplx_fact[dt] = []
                    for i in range(len(sbk_mass)):
                        cmplx_fact[dt].append([])
                        for j in range(3):
                            cmplx_fact[dt][-1].append((sbk_mass[i][j]/abs(cond_mass[i][j][j]))**1.5)
                
        else:
            sbk_mass = self.get_seebeck_eff_mass(output, temp, False, Lambda)
            cond_mass = self.get_average_eff_mass(output=output, doping_levels=False)[temp]
            
            if output == 'average':
                cmplx_fact = [ (m_s/abs(m_c))**1.5 for m_s,m_c in zip(sbk_mass,cond_mass)]
            elif output == 'tensor':
                cmplx_fact = []
                for i in range(len(sbk_mass)):
                    cmplx_fact.append([])
                    for j in range(3):
                        cmplx_fact[-1].append((sbk_mass[i][j]/abs(cond_mass[i][j][j]))**1.5)
            
        return cmplx_fact

    def get_extreme(self, target_prop, maximize=True, min_temp=None,
                    max_temp=None, min_doping=None, max_doping=None,
                    isotropy_tolerance=0.05, use_average=True):

        """
        This method takes in eigenvalues over a range of carriers,
        temperatures, and doping levels, and tells you what is the "best"
        value that can be achieved for the given target_property. Note that
        this method searches the doping dict only, not the full mu dict.

        Args:
            target_prop: target property, i.e. "seebeck", "power factor",
                         "conductivity", "kappa", or "zt"
            maximize: True to maximize, False to minimize (e.g. kappa)
            min_temp: minimum temperature allowed
            max_temp: maximum temperature allowed
            min_doping: minimum doping allowed (e.g., 1E18)
            max_doping: maximum doping allowed (e.g., 1E20)
            isotropy_tolerance: tolerance for isotropic (0.05 = 5%)
            use_average: True for avg of eigenval, False for max eigenval

        Returns:
            A dictionary with keys {"p", "n", "best"} with sub-keys:
            {"value", "temperature", "doping", "isotropic"}

        """

        def is_isotropic(x, isotropy_tolerance):
            """
            Internal method to tell you if 3-vector "x" is isotropic

            Args:
                x: the vector to determine isotropy for
                isotropy_tolerance: tolerance, e.g. 0.05 is 5%
            """
            if len(x) != 3:
                raise ValueError("Invalid input to is_isotropic!")

            st = sorted(x)
            return bool(all([st[0], st[1], st[2]]) and \
                        (abs((st[1] - st[0]) / st[1]) <= isotropy_tolerance) and \
                        (abs((st[2] - st[0])) / st[2] <= isotropy_tolerance) and \
                        (abs((st[2] - st[1]) / st[2]) <= isotropy_tolerance))

        if target_prop.lower() == "seebeck":
            d = self.get_seebeck(output="eigs", doping_levels=True)

        elif target_prop.lower() == "power factor":
            d = self.get_power_factor(output="eigs", doping_levels=True)

        elif target_prop.lower() == "conductivity":
            d = self.get_conductivity(output="eigs", doping_levels=True)

        elif target_prop.lower() == "kappa":
            d = self.get_thermal_conductivity(output="eigs",
                                              doping_levels=True)
        elif target_prop.lower() == "zt":
            d = self.get_zt(output="eigs", doping_levels=True)

        else:
            raise ValueError("Target property: {} not recognized!".
                             format(target_prop))

        absval = True  # take the absolute value of properties

        x_val = None
        x_temp = None
        x_doping = None
        x_isotropic = None
        output = {}

        min_temp = min_temp or 0
        max_temp = max_temp or float('inf')
        min_doping = min_doping or 0
        max_doping = max_doping or float('inf')

        for pn in ('p', 'n'):
            for t in d[pn]:  # temperatures
                if min_temp <= float(t) <= max_temp:
                    for didx, evs in enumerate(d[pn][t]):
                        doping_lvl = self.doping[pn][didx]
                        if min_doping <= doping_lvl <= max_doping:
                            isotropic = is_isotropic(evs, isotropy_tolerance)
                            if absval:
                                evs = [abs(x) for x in evs]
                            if use_average:
                                val = float(sum(evs)) / len(evs)
                            else:
                                val = max(evs)
                            if x_val is None or (val > x_val and maximize) \
                                    or (val < x_val and not maximize):
                                x_val = val
                                x_temp = t
                                x_doping = doping_lvl
                                x_isotropic = isotropic

            output[pn] = {'value': x_val, 'temperature': x_temp,
                          'doping': x_doping, 'isotropic': x_isotropic}
            x_val = None

        if maximize:
            max_type = 'p' if output['p']['value'] >= \
                              output['n']['value'] else 'n'
        else:
            max_type = 'p' if output['p']['value'] <= \
                              output['n']['value'] else 'n'

        output['best'] = output[max_type]
        output['best']['carrier_type'] = max_type

        return output

    @staticmethod
    def _format_to_output(tensor, tensor_doping, output, doping_levels,
                          multi=1.0):
        if doping_levels:
            full_tensor = tensor_doping
            result = {doping: {t: [] for t in tensor_doping[doping]} for doping
                      in tensor_doping}
            for doping in full_tensor:
                for temp in full_tensor[doping]:
                    for i in range(len(full_tensor[doping][temp])):
                        if output in ['eig', 'eigs']:
                            result[doping][temp].append(sorted(
                                np.linalg.eigh(full_tensor[doping][temp][i])[
                                    0] * multi))
                        elif output == 'tensor':
                            result[doping][temp].append(
                                np.array(full_tensor[doping][temp][i]) * multi)
                        elif output == 'average':
                            result[doping][temp].append(
                                (full_tensor[doping][temp][i][0][0] \
                                 + full_tensor[doping][temp][i][1][1] \
                                 + full_tensor[doping][temp][i][2][
                                     2]) * multi / 3.0)
                        else:
                            raise ValueError("Unknown output format: "
                                             "{}".format(output))
        else:
            full_tensor = tensor
            result = {t: [] for t in tensor}
            for temp in full_tensor:
                for i in range(len(tensor[temp])):
                    if output in ['eig', 'eigs']:
                        result[temp].append(sorted(
                            np.linalg.eigh(full_tensor[temp][i])[0] * multi))
                    elif output == 'tensor':
                        result[temp].append(
                            np.array(full_tensor[temp][i]) * multi)
                    elif output == 'average':
                        result[temp].append((full_tensor[temp][i][0][0]
                                             + full_tensor[temp][i][1][1]
                                             + full_tensor[temp][i][2][
                                                 2]) * multi / 3.0)
                    else:
                        raise ValueError("Unknown output format: {}".
                                         format(output))
        return result

    def get_complete_dos(self, structure, analyzer_for_second_spin=None):
        """
            Gives a CompleteDos object with the DOS from the interpolated
            projected band structure
            Args:
                the structure (necessary to identify sites for projection)
                analyzer_for_second_spin must be specified to have a
                CompleteDos with both Spin components
            Returns:
                a CompleteDos object
            Example of use in case of spin polarized case:

                BoltztrapRunner(bs=bs,nelec=10,run_type="DOS",spin=1).run(path_dir='dos_up/')
                an_up=BoltztrapAnalyzer.from_files("dos_up/boltztrap/",dos_spin=1)

                BoltztrapRunner(bs=bs,nelec=10,run_type="DOS",spin=-1).run(path_dir='dos_dw/')
                an_dw=BoltztrapAnalyzer.from_files("dos_dw/boltztrap/",dos_spin=-1)

                cdos=an_up.get_complete_dos(bs.structure,an_dw)

        """
        pdoss = {}
        spin_1 = list(self.dos.densities.keys())[0]

        if analyzer_for_second_spin:
            if not np.all(self.dos.energies ==
                                  analyzer_for_second_spin.dos.energies):
                raise BoltztrapError(
                    "Dos merging error: energies of the two dos are different")

            spin_2 = list(analyzer_for_second_spin.dos.densities.keys())[0]
            if spin_1 == spin_2:
                raise BoltztrapError(
                    "Dos merging error: spin component are the same")

        for s in self._dos_partial:
            if structure.sites[int(s)] not in pdoss:
                pdoss[structure.sites[int(s)]] = {}
            for o in self._dos_partial[s]:
                if Orbital[o] not in pdoss[structure.sites[int(s)]]:
                    pdoss[structure.sites[int(s)]][Orbital[o]] = {}
                pdoss[structure.sites[int(s)]][Orbital[o]][
                    spin_1] = self._dos_partial[s][o]
                if analyzer_for_second_spin:
                    pdoss[structure.sites[int(s)]][Orbital[o]][
                        spin_2] = analyzer_for_second_spin._dos_partial[s][o]
        if analyzer_for_second_spin:
            tdos = Dos(self.dos.efermi, self.dos.energies,
                       {spin_1: self.dos.densities[spin_1],
                        spin_2: analyzer_for_second_spin.dos.densities[
                            spin_2]})
        else:
            tdos = self.dos

        return CompleteDos(structure, total_dos=tdos, pdoss=pdoss)

    def get_mu_bounds(self, temp=300):
        return min(self.mu_doping['p'][temp]), max(self.mu_doping['n'][temp])

    def get_carrier_concentration(self):
        """
        gives the carrier concentration (in cm^-3)

        Returns
            a dictionary {temp:[]} with an array of carrier concentration
            (in cm^-3) at each temperature
            The array relates to each step of electron chemical potential
        """

        return {temp: [1e24 * i / self.vol for i in self._carrier_conc[temp]]
                for temp in self._carrier_conc}

    def get_hall_carrier_concentration(self):
        """
        gives the Hall carrier concentration (in cm^-3). This is the trace of
        the Hall tensor (see Boltztrap source code) Hall carrier concentration
        are not always exactly the same than carrier concentration.

        Returns
            a dictionary {temp:[]} with an array of Hall carrier concentration
            (in cm^-3) at each temperature The array relates to each step of
            electron chemical potential
        """
        result = {temp: [] for temp in self._hall}
        for temp in self._hall:
            for i in self._hall[temp]:
                trace = (i[1][2][0] + i[2][0][1] + i[0][1][2]) / 3.0
                if trace != 0.0:
                    result[temp].append(1e-6 / (trace * constants.e))
                else:
                    result[temp].append(0.0)
        return result

    @staticmethod
    def parse_outputtrans(path_dir):
        """
        Parses .outputtrans file

        Args:
            path_dir: dir containing boltztrap.outputtrans

        Returns:
            tuple - (run_type, warning, efermi, gap, doping_levels)

        """
        run_type = None
        warning = None
        efermi = None
        gap = None
        doping_levels = []

        with open(os.path.join(path_dir, "boltztrap.outputtrans"), 'r') \
                as f:
            for line in f:
                if "WARNING" in line:
                    warning = line
                elif "Calc type:" in line:
                    run_type = line.split()[-1]
                elif line.startswith("VBM"):
                    efermi = Energy(line.split()[1], "Ry").to("eV")
                elif line.startswith("Egap:"):
                    gap = Energy(float(line.split()[1]), "Ry").to("eV")
                elif line.startswith("Doping level number"):
                    doping_levels.append(float(line.split()[6]))

        return run_type, warning, efermi, gap, doping_levels

    @staticmethod
    def parse_transdos(path_dir, efermi, dos_spin=1, trim_dos=False):

        """
        Parses .transdos (total DOS) and .transdos_x_y (partial DOS) files
        Args:
            path_dir: (str) dir containing DOS files
            efermi: (float) Fermi energy
            dos_spin: (int) -1 for spin down, +1 for spin up
            trim_dos: (bool) whether to post-process / trim DOS

        Returns:
            tuple - (DOS, dict of partial DOS)
        """

        data_dos = {'total': [], 'partial': {}}
        # parse the total DOS data
        ## format is energy, DOS, integrated DOS
        with open(os.path.join(path_dir, "boltztrap.transdos"), 'r') as f:
            count_series = 0  # TODO: why is count_series needed?
            for line in f:
                if line.lstrip().startswith("#"):
                    count_series += 1
                    if count_series > 1:
                        break
                else:
                    data_dos['total'].append(
                        [Energy(float(line.split()[0]), "Ry").to("eV"),
                         float(line.split()[1])])
                    total_elec = float(line.split()[2])

        lw_l = 0
        hg_l = -len(data_dos['total'])
        if trim_dos:
            # Francesco knows what this does
            # It has something to do with a trick of adding fake energies
            # at the endpoints of the DOS, and then re-trimming it. This is
            # to get the same energy scale for up and down spin DOS.
            tmp_data = np.array(data_dos['total'])
            tmp_den = np.trim_zeros(tmp_data[:, 1], 'f')[1:]
            lw_l = len(tmp_data[:, 1]) - len(tmp_den)
            tmp_ene = tmp_data[lw_l:, 0]
            tmp_den = np.trim_zeros(tmp_den, 'b')[:-1]
            hg_l = len(tmp_ene) - len(tmp_den)
            tmp_ene = tmp_ene[:-hg_l]
            tmp_data = np.vstack((tmp_ene, tmp_den)).T
            data_dos['total'] = tmp_data.tolist()

        # parse partial DOS data
        for file_name in os.listdir(path_dir):
            if file_name.endswith(
                    "transdos") and file_name != 'boltztrap.transdos':
                tokens = file_name.split(".")[1].split("_")
                site = tokens[1]
                orb = '_'.join(tokens[2:])
                with open(os.path.join(path_dir, file_name), 'r') as f:
                    for line in f:
                        if not line.lstrip().startswith(" #"):
                            if site not in data_dos['partial']:
                                data_dos['partial'][site] = {}
                            if orb not in data_dos['partial'][site]:
                                data_dos['partial'][site][orb] = []
                            data_dos['partial'][site][orb].append(
                                float(line.split()[1]))
                data_dos['partial'][site][orb] = data_dos['partial'][site][
                                                     orb][lw_l:-hg_l]

        dos_full = {'energy': [], 'density': []}

        for t in data_dos['total']:
            dos_full['energy'].append(t[0])
            dos_full['density'].append(t[1])

        dos = Dos(efermi, dos_full['energy'],
                  {Spin(dos_spin): dos_full['density']})
        dos_partial = data_dos['partial']  # TODO: make this real DOS object?

        return dos, dos_partial

    @staticmethod
    def parse_intrans(path_dir):
        """
        Parses boltztrap.intrans mainly to extract the value of scissor applied to the bands or some other inputs
        Args:
            path_dir: (str) dir containing the boltztrap.intrans file
        Returns:
            intrans (dict): a dictionary containing various inputs that had been used in the Boltztrap run.
        """
        intrans = {}
        with open(os.path.join(path_dir, "boltztrap.intrans"), 'r') as f:
            for line in f:
                if "iskip" in line:
                    intrans["scissor"] = Energy(float(line.split(" ")[3]),
                                                "Ry").to("eV")
                if "HISTO" in line or "TETRA" in line:
                    intrans["dos_type"] = line[:-1]
        return intrans

    @staticmethod
    def parse_struct(path_dir):
        """
        Parses boltztrap.struct file (only the volume)
        Args:
            path_dir: (str) dir containing the boltztrap.struct file

        Returns:
            (float) volume
        """
        with open(os.path.join(path_dir, "boltztrap.struct"), 'r') as f:
            tokens = f.readlines()
            return Lattice([[Length(float(tokens[i].split()[j]), "bohr").
                           to("ang") for j in range(3)] for i in
                            range(1, 4)]).volume

    @staticmethod
    def parse_cond_and_hall(path_dir, doping_levels=None):
        """
        Parses the conductivity and Hall tensors
        Args:
            path_dir: Path containing .condtens / .halltens files
            doping_levels: ([float]) - doping lvls, parse outtrans to get this

        Returns:
            mu_steps, cond, seebeck, kappa, hall, pn_doping_levels,
            mu_doping, seebeck_doping, cond_doping, kappa_doping,
            hall_doping, carrier_conc
        """

        # Step 1: parse raw data but do not convert to final format
        t_steps = set()
        mu_steps = set()
        data_full = []
        data_hall = []
        data_doping_full = []
        data_doping_hall = []
        doping_levels = doping_levels or []

        # parse the full conductivity/Seebeck/kappa0/etc data
        ## also initialize t_steps and mu_steps
        with open(os.path.join(path_dir, "boltztrap.condtens"), 'r') as f:
            for line in f:
                if not line.startswith("#"):
                    mu_steps.add(float(line.split()[0]))
                    t_steps.add(int(float(line.split()[1])))
                    data_full.append([float(c) for c in line.split()])

        # parse the full Hall tensor
        with open(os.path.join(path_dir, "boltztrap.halltens"), 'r') as f:
            for line in f:
                if not line.startswith("#"):
                    data_hall.append([float(c) for c in line.split()])

        if len(doping_levels) != 0:
            # parse doping levels version of full cond. tensor, etc.
            with open(
                    os.path.join(path_dir, "boltztrap.condtens_fixdoping"),
                    'r') as f:
                for line in f:
                    if not line.startswith("#") and len(line) > 2:
                        data_doping_full.append([float(c)
                                                 for c in line.split()])

            # parse doping levels version of full hall tensor
            with open(
                    os.path.join(path_dir, "boltztrap.halltens_fixdoping"),
                    'r') as f:
                for line in f:
                    if not line.startswith("#") and len(line) > 2:
                        data_doping_hall.append(
                            [float(c) for c in line.split()])

        # Step 2: convert raw data to final format

        # sort t and mu_steps (b/c they are sets not lists)
        # and convert to correct energy
        t_steps = sorted([t for t in t_steps])
        mu_steps = sorted([Energy(m, "Ry").to("eV") for m in mu_steps])

        # initialize output variables - could use defaultdict instead
        # I am leaving things like this for clarity
        cond = {t: [] for t in t_steps}
        seebeck = {t: [] for t in t_steps}
        kappa = {t: [] for t in t_steps}
        hall = {t: [] for t in t_steps}
        carrier_conc = {t: [] for t in t_steps}
        dos_full = {'energy': [], 'density': []}

        mu_doping = {'p': {t: [] for t in t_steps},
                     'n': {t: [] for t in t_steps}}
        seebeck_doping = {'p': {t: [] for t in t_steps},
                          'n': {t: [] for t in t_steps}}
        cond_doping = {'p': {t: [] for t in t_steps},
                       'n': {t: [] for t in t_steps}}
        kappa_doping = {'p': {t: [] for t in t_steps},
                        'n': {t: [] for t in t_steps}}
        hall_doping = {'p': {t: [] for t in t_steps},
                       'n': {t: [] for t in t_steps}}

        # process doping levels
        pn_doping_levels = {'p': [], 'n': []}
        for d in doping_levels:
            if d > 0:
                pn_doping_levels['p'].append(d)
            else:
                pn_doping_levels['n'].append(-d)

        # process raw conductivity data, etc.
        for d in data_full:
            temp, doping = d[1], d[2]
            carrier_conc[temp].append(doping)

            cond[temp].append(np.reshape(d[3:12], (3, 3)).tolist())
            seebeck[temp].append(np.reshape(d[12:21], (3, 3)).tolist())
            kappa[temp].append(np.reshape(d[21:30], (3, 3)).tolist())

        # process raw Hall data
        for d in data_hall:
            temp, doping = d[1], d[2]
            hall_tens = [np.reshape(d[3:12], (3, 3)).tolist(),
                         np.reshape(d[12:21], (3, 3)).tolist(),
                         np.reshape(d[21:30], (3, 3)).tolist()]
            hall[temp].append(hall_tens)

        # process doping conductivity data, etc.
        for d in data_doping_full:
            temp, doping, mu = d[0], d[1], d[-1]
            pn = 'p' if doping > 0 else 'n'
            mu_doping[pn][temp].append(Energy(mu, "Ry").to("eV"))
            cond_doping[pn][temp].append(
                np.reshape(d[2:11], (3, 3)).tolist())
            seebeck_doping[pn][temp].append(
                np.reshape(d[11:20], (3, 3)).tolist())
            kappa_doping[pn][temp].append(
                np.reshape(d[20:29], (3, 3)).tolist())

        # process doping Hall data
        for d in data_doping_hall:
            temp, doping, mu = d[0], d[1], d[-1]
            pn = 'p' if doping > 0 else 'n'
            hall_tens = [np.reshape(d[2:11], (3, 3)).tolist(),
                         np.reshape(d[11:20], (3, 3)).tolist(),
                         np.reshape(d[20:29], (3, 3)).tolist()]
            hall_doping[pn][temp].append(hall_tens)

        return mu_steps, cond, seebeck, kappa, hall, pn_doping_levels, \
               mu_doping, seebeck_doping, cond_doping, kappa_doping, \
               hall_doping, carrier_conc

    @staticmethod
    def from_files(path_dir, dos_spin=1):
        """
        get a BoltztrapAnalyzer object from a set of files

        Args:
            path_dir: directory where the boltztrap files are
            dos_spin: in DOS mode, set to 1 for spin up and -1 for spin down

        Returns:
            a BoltztrapAnalyzer object

        """
        run_type, warning, efermi, gap, doping_levels = \
            BoltztrapAnalyzer.parse_outputtrans(path_dir)

        vol = BoltztrapAnalyzer.parse_struct(path_dir)

        intrans = BoltztrapAnalyzer.parse_intrans(path_dir)

        if run_type == "BOLTZ":
            dos, pdos = BoltztrapAnalyzer.parse_transdos(
                path_dir, efermi, dos_spin=dos_spin, trim_dos=False)

            mu_steps, cond, seebeck, kappa, hall, pn_doping_levels, mu_doping, \
            seebeck_doping, cond_doping, kappa_doping, hall_doping, \
            carrier_conc = BoltztrapAnalyzer. \
                parse_cond_and_hall(path_dir, doping_levels)

            return BoltztrapAnalyzer(
                gap, mu_steps, cond, seebeck, kappa, hall, pn_doping_levels,
                mu_doping, seebeck_doping, cond_doping, kappa_doping,
                hall_doping, intrans, dos, pdos, carrier_conc, vol, warning)

        elif run_type == "DOS":
            trim = True if intrans["dos_type"] == "HISTO" else False
            dos, pdos = BoltztrapAnalyzer.parse_transdos(
                path_dir, efermi, dos_spin=dos_spin, trim_dos=trim)

            return BoltztrapAnalyzer(gap=gap, dos=dos, dos_partial=pdos,
                                     warning=warning, vol=vol)

        elif run_type == "BANDS":
            bz_kpoints = np.loadtxt(
                os.path.join(path_dir, "boltztrap_band.dat"))[:, -3:]
            bz_bands = np.loadtxt(
                os.path.join(path_dir, "boltztrap_band.dat"))[:, 1:-6]
            return BoltztrapAnalyzer(bz_bands=bz_bands, bz_kpoints=bz_kpoints,
                                     warning=warning, vol=vol)

        elif run_type == "FERMI":
            """
            """

            if os.path.exists(os.path.join(path_dir, 'boltztrap_BZ.cube')):
                fs_data = read_cube_file(
                    os.path.join(path_dir, 'boltztrap_BZ.cube'))
            elif os.path.exists(os.path.join(path_dir, 'fort.30')):
                fs_data = read_cube_file(os.path.join(path_dir, 'fort.30'))
            else:
                raise BoltztrapError("No data file found for fermi surface")
            return BoltztrapAnalyzer(fermi_surface_data=fs_data)

        else:
            raise ValueError("Run type: {} not recognized!".format(run_type))

    def as_dict(self):

        results = {'gap': self.gap,
                   'mu_steps': self.mu_steps,
                   'intrans': self.intrans,
                   'cond': self._cond,
                   'seebeck': self._seebeck,
                   'kappa': self._kappa,
                   'hall': self._hall,
                   'doping': self.doping,
                   'mu_doping': self.mu_doping,
                   'seebeck_doping': self._seebeck_doping,
                   'cond_doping': self._cond_doping,
                   'kappa_doping': self._kappa_doping,
                   'hall_doping': self._hall_doping,
                   'dos': self.dos.as_dict(),
                   'dos_partial': self._dos_partial,
                   'carrier_conc': self._carrier_conc,
                   'vol': self.vol,
                   'warning': self.warning}
        return jsanitize(results)

    @staticmethod
    def from_dict(data):
        def _make_float_array(a):
            res = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]
            for i in range(3):
                for j in range(3):
                    res[i][j] = float(a[i][j])
            return res

        def _make_float_hall(a):
            return [i for i in a[:27]]
        
        intrans = data.get('intrans')
        gap = data.get('gap')
        mu_steps = [float(d) for d in data['mu_steps']] if \
            'mu_steps' in data else None
        cond = {int(d): [_make_float_array(v) for v in data['cond'][d]]
                for d in data['cond']} if 'cond' in data else None
        seebeck = {int(d): [_make_float_array(v) for v in data['seebeck'][d]]
                   for d in data['seebeck']} if 'seebeck' in data else None
        kappa = {int(d): [_make_float_array(v) for v in data['kappa'][d]]
                 for d in data['kappa']} if 'kappa' in data else None
        hall = {int(d): [_make_float_hall(v) for v in data['hall'][d]]
                for d in data['hall']} if 'hall' in data else None
        doping = {'p': [float(d) for d in data['doping']['p']],
                  'n': [float(d) for d in data['doping']['n']]} if \
            'doping' in data else None

        mu_doping = {'p': {int(d): [
            float(v) for v in data['mu_doping']['p'][d]] for d in
                           data['mu_doping']['p']}, 'n':
                         {int(d): [float(v) for v in data['mu_doping']['n'][d]]
                          for d in data['mu_doping'][
                              'n']}} if 'mu_doping' in data else None

        seebeck_doping = {'p': {int(d): [
            _make_float_array(v) for v in data['seebeck_doping']['p'][d]]
                                for d in data['seebeck_doping']['p']}, 'n':
                              {int(d): [_make_float_array(v) for v in
                                        data['seebeck_doping']['n'][d]] for d in
                               data['seebeck_doping'][
                                   'n']}} if 'seebeck_doping' in data \
            else None

        cond_doping = {'p': {int(d): [_make_float_array(v)
                                      for v in data['cond_doping']['p'][d]]
                             for d in data['cond_doping']['p']}, 'n':
                           {int(d): [_make_float_array(v) for v in
                                     data['cond_doping']['n'][d]] for
                            d in data['cond_doping'][
                                'n']}} if 'cond_doping' in data else None

        kappa_doping = {'p': {int(d): [_make_float_array(v)
                                       for v in data['kappa_doping']['p'][d]]
                              for d in data['kappa_doping']['p']},
                        'n': {int(d): [_make_float_array(v) for v in
                                       data['kappa_doping']['n'][d]]
                              for d in data['kappa_doping']['n']}} \
            if 'kappa_doping' in data else None

        hall_doping = {'p': {int(d): [_make_float_hall(v) for v in
                                      data['hall_doping']['p'][d]] for d in
                             data['hall_doping']['p']}, 'n':
                           {int(d): [_make_float_hall(v) for v in
                                     data['hall_doping']['n'][d]] for d in
                            data['hall_doping'][
                                'n']}} if "hall_doping" in data else None

        dos = Dos.from_dict(data['dos']) if 'dos' in data else None
        dos_partial = data.get('dos_partial')
        carrier_conc = data.get('carrier_conc')
        vol = data.get('vol')
        warning = data.get('warning')

        return BoltztrapAnalyzer(gap=gap,
                                 mu_steps=mu_steps,
                                 cond=cond,
                                 seebeck=seebeck,
                                 kappa=kappa,
                                 hall=hall,
                                 doping=doping,
                                 mu_doping=mu_doping,
                                 seebeck_doping=seebeck_doping,
                                 cond_doping=cond_doping,
                                 kappa_doping=kappa_doping,
                                 hall_doping=hall_doping,
                                 dos=dos,
                                 dos_partial=dos_partial,
                                 carrier_conc=carrier_conc,
                                 vol=vol,
                                 warning=warning)


def read_cube_file(filename):
    with open(filename, 'rt') as f:
            natoms = 0
            count_line = 0
            for line in f:
                line = line.rstrip("\n")
                if count_line == 0 and "CUBE" not in line:
                    raise ValueError("CUBE file format not recognized")

                if count_line == 2:
                    tokens = line.split()
                    origin = [float(tokens[i]) for i in range(1,4)]
                    natoms = int(tokens[0])
                if count_line == 3:
                    tokens = line.split()
                    a1 = [float(tokens[i]) for i in range(1,4)]
                    n1 = int(tokens[0])
                elif count_line == 4:
                    tokens = line.split()
                    a2 = [float(tokens[i]) for i in range(1,4)]
                    n2 = int(tokens[0])
                elif count_line == 5:
                    tokens = line.split()
                    a3 = [float(tokens[i]) for i in range(1,4)]
                    n3 = int(tokens[0])
                    #kpoints=[[[0 for i in range(0,n1)] for j in range(0,n2)] for l in range(0,n3)]
                elif count_line > 5:
                    break

                count_line += 1

    if 'fort.30' in filename:
        energy_data = np.genfromtxt(filename,skip_header=natoms+6,skip_footer=1)
        nlines_data = len(energy_data)
        last_line = np.genfromtxt(filename,skip_header=nlines_data+natoms+6)
        energy_data = np.append(energy_data.flatten(),last_line).reshape(n1,n2,n3)
    elif 'boltztrap_BZ.cube' in filename:
        energy_data = np.loadtxt(filename,skiprows=natoms+6).reshape(n1,n2,n3)

    energy_data /= Energy(1, "eV").to("Ry")

    return energy_data


def compare_sym_bands(bands_obj, bands_ref_obj, nb=None):
    """
    Compute the mean of correlation between bzt and vasp bandstructure on
    sym line, for all bands and locally (for each branches) the difference
    squared (%) if nb is specified.
    """

    nkpt = len(bands_obj.kpoints)
    if bands_ref_obj.is_spin_polarized:
        nbands = min(bands_obj.nb_bands, 2 * bands_ref_obj.nb_bands)
    else:
        # TODO: why is this needed? Shouldn't pmg take care of nb_bands?
        nbands = min(len(bands_obj.bands[Spin.up]),
                     len(bands_ref_obj.bands[Spin.up]))
    # print(nbands)
    arr_bands = np.array(bands_obj.bands[Spin.up][:nbands])
    # arr_bands_lavg = (arr_bands-np.mean(arr_bands,axis=1).reshape(nbands,1))

    if bands_ref_obj.is_spin_polarized:
        arr_bands_ref_up = np.array(bands_ref_obj.bands[Spin.up])
        arr_bands_ref_dw = np.array(bands_ref_obj.bands[Spin.down])
        # print(arr_bands_ref_up.shape)
        arr_bands_ref = np.vstack((arr_bands_ref_up, arr_bands_ref_dw))
        arr_bands_ref = np.sort(arr_bands_ref, axis=0)[:nbands]
        # print(arr_bands_ref.shape)
    else:
        arr_bands_ref = np.array(bands_ref_obj.bands[Spin.up][:nbands])

    # arr_bands_ref_lavg =
    # (arr_bands_ref-np.mean(arr_bands_ref,axis=1).reshape(nbands,1))

    # err = np.sum((arr_bands_lavg-arr_bands_ref_lavg)**2,axis=1)/nkpt
    corr = np.array(
        [distance.correlation(arr_bands[idx], arr_bands_ref[idx]) for idx in
         range(nbands)])

    if type(nb) == int: nb = [nb]

    bcheck = {}

    if max(nb) < nbands:
        branches = [[s['start_index'], s['end_index'], s['name']] for s in
                    bands_ref_obj.branches]

        if not bands_obj.is_metal() and not bands_ref_obj.is_metal():
            zero_ref = bands_ref_obj.get_vbm()['energy']
            zero = bands_obj.get_vbm()['energy']
            if not zero:
                vbm = bands_ref_obj.get_vbm()['band_index'][Spin.up][-1]
                zero = max(arr_bands[vbm])
        else:
            zero_ref = 0  # bands_ref_obj.efermi
            zero = 0  # bands_obj.efermi
            print(zero, zero_ref)

        for nbi in nb:
            bcheck[nbi] = {}

            bcheck[nbi]['Dist'] = np.mean(abs(arr_bands[nbi] - zero
                                              - arr_bands_ref[nbi] + zero_ref))
            bcheck[nbi]['Corr'] = corr[nbi]

            for start, end, name in branches:
                # werr.append((sum((arr_bands_corr[nb][start:end+1] -
                # arr_bands_ref_corr[nb][start:end+1])**2)/(end+1-start)*100,name))
                bcheck[nbi][name] = np.mean(abs(arr_bands[nbi][start:end + 1]
                                                - zero
                                                - arr_bands_ref[nbi][
                                                  start:end + 1] + zero_ref))
    else:
        bcheck = "No nb given"

    return bcheck


def seebeck_spb(eta,Lambda=0.5):
    """
        Seebeck analytic formula in the single parabolic model
    """
    from fdint import fdk
    
    return constants.k/constants.e * ((2. + Lambda) * fdk( 1.+ Lambda, eta)/ 
                 ((1.+Lambda)*fdk(Lambda, eta))- eta) * 1e+6

def eta_from_seebeck(seeb,Lambda):
    """ 
        It takes a value of seebeck and adjusts the analytic seebeck until it's equal
        Returns: eta where the two seebeck coefficients are equal 
        (reduced chemical potential)
    """
    from scipy.optimize import fsolve
    out = fsolve(lambda x: (seebeck_spb(x,Lambda) - abs(seeb)) ** 2, 1.,full_output=True)
    return out[0][0]


def seebeck_eff_mass_from_carr(eta, n, T, Lambda):
    """
        Calculate seebeck effective mass at a certain carrier concentration
        eta in kB*T units, n in cm-3, T in K, returns mass in m0 units
    """
    from fdint import fdk

    return (2 * np.pi**2 * abs(n) * 10 ** 6 / (fdk(0.5,eta))) ** (2. / 3)\
            / (2 * constants.m_e * constants.k * T / (constants.h/2/np.pi) ** 2)


def seebeck_eff_mass_from_seebeck_carr(seeb, n, T, Lambda):
    """
        Find the chemical potential where analytic and calculated seebeck are identical
        and then calculate the seebeck effective mass at that chemical potential and 
        a certain carrier concentration n
    """
    try:
        from fdint import fdk
    except ImportError:
        raise BoltztrapError("fdint module not found. Please, install it.\n"+
                            "It is needed to calculate Fermi integral quickly.")

    eta = eta_from_seebeck(seeb,Lambda)
    mass = seebeck_eff_mass_from_carr(eta, n, T, Lambda)
    return mass



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.


"""
This module implements abstract base classes for post-processing entries.
Any class which modifies entries should inherit these classes.
"""

__author__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2011, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyuep@gmail.com"
__date__ = "Oct 6, 2011"

import abc


class EntryPostProcessor(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def process_entry(self, entry):
        """
        Process a single entry.

        Args:
            entry: An ComputedEntry object.

        Returns:
            An processed entry. None if entry is not compatible within the
            processing scheme.
        """
        return

    @abc.abstractmethod
    def process_entries(self, entries):
        """
        Process a sequence of entries.

        Args:
            entries: A sequence of ComputedEntries.

        Returns:
            An list of processed entries.  ComputedEntries in the original list
            which are not compatible with the processing scheme are excluded.
        """
        return

    @property
    @abc.abstractmethod
    def corrected_compound_formulas(self):
        """
        List of compound formulas that are corrected.
        """
        return



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

__author__ = 'Anubhav Jain'
__copyright__ = 'Copyright 2014, The Materials Project'
__version__ = '0.1'
__maintainer__ = 'Anubhav Jain'
__email__ = 'ajain@lbl.gov'
__date__ = 'Oct 03, 2014'


# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.


import logging
import numpy as np
import time
from pymatgen.core.structure import Structure
from pymatgen.core.sites import PeriodicSite
from monty.json import MSONable
from scipy.spatial import Voronoi

from pymatgen.analysis.chemenv.utils.coordination_geometry_utils import my_solid_angle
from pymatgen.analysis.chemenv.utils.coordination_geometry_utils import get_lower_and_upper_f
from pymatgen.analysis.chemenv.utils.coordination_geometry_utils import rectangle_surface_intersection
from pymatgen.analysis.chemenv.utils.defs_utils import AdditionalConditions
from pymatgen.analysis.chemenv.utils.math_utils import normal_cdf_step

"""
This module contains the object used to describe the possible bonded atoms based on a Voronoi analysis
"""

__author__ = "David Waroquiers"
__copyright__ = "Copyright 2012, The Materials Project"
__credits__ = "Geoffroy Hautier"
__version__ = "2.0"
__maintainer__ = "David Waroquiers"
__email__ = "david.waroquiers@gmail.com"
__date__ = "Feb 20, 2016"


def from_bson_voronoi_list(bson_nb_voro_list, structure):
    """
    Returns the voronoi_list needed for the VoronoiContainer object from a bson-encoded voronoi_list (composed of
    vlist and bson_nb_voro_list).
    :param vlist: List of voronoi objects
    :param bson_nb_voro_list: List of periodic sites involved in the Voronoi
    :return: The voronoi_list needed for the VoronoiContainer (with PeriodicSites as keys of the dictionary - not
    allowed in the BSON format)
    """
    voronoi_list = [None] * len(bson_nb_voro_list)
    for isite, voro in enumerate(bson_nb_voro_list):
        if voro is None or voro == 'None':
            continue
        voronoi_list[isite] = []
        for psd, dd in voro:
            struct_site = structure[dd['index']]
            periodic_site = PeriodicSite(struct_site._species, struct_site.frac_coords + psd[1],
                                         struct_site._lattice, properties=struct_site.properties)
            voronoi_list[isite].append((periodic_site, dd))
    return voronoi_list

def from_bson_voronoi_list2(bson_nb_voro_list2, structure):
    """
    Returns the voronoi_list needed for the VoronoiContainer object from a bson-encoded voronoi_list (composed of
    vlist and bson_nb_voro_list).
    :param vlist: List of voronoi objects
    :param bson_nb_voro_list: List of periodic sites involved in the Voronoi
    :return: The voronoi_list needed for the VoronoiContainer (with PeriodicSites as keys of the dictionary - not
    allowed in the BSON format)
    """
    voronoi_list = [None] * len(bson_nb_voro_list2)
    for isite, voro in enumerate(bson_nb_voro_list2):
        if voro is None or voro == 'None':
            continue
        voronoi_list[isite] = []
        for psd, dd in voro:
            struct_site = structure[dd['index']]
            periodic_site = PeriodicSite(struct_site._species, struct_site.frac_coords + psd[1],
                                         struct_site._lattice, properties=struct_site.properties)
            dd['site'] = periodic_site
            voronoi_list[isite].append(dd)
    return voronoi_list


class DetailedVoronoiContainer(MSONable):
    """
    Class used to store the full Voronoi of a given structure.
    """
    AC = AdditionalConditions()
    default_voronoi_cutoff = 10.0
    default_normalized_distance_tolerance = 1e-5
    default_normalized_angle_tolerance = 1e-3

    def __init__(self, structure=None, voronoi_list=None, voronoi_list2=None,
                 voronoi_cutoff=default_voronoi_cutoff, isites=None,
                 normalized_distance_tolerance=default_normalized_distance_tolerance,
                 normalized_angle_tolerance=default_normalized_angle_tolerance,
                 additional_conditions=None, valences=None,
                 maximum_distance_factor=None, minimum_angle_factor=None):
        """
        Constructor for the VoronoiContainer object. Either a structure is given, in which case the Voronoi is
        computed, or the different components of the VoronoiContainer are given (used in the from_dict method)
        :param structure: Structure for which the Voronoi is computed
        :param voronoi_list: List of voronoi polyhedrons for each site
        :param voronoi_cutoff: cutoff used for the voronoi
        :param isites: indices of sites for which the Voronoi has to be computed
        :raise: RuntimeError if the Voronoi cannot be constructed
        """
        self.normalized_distance_tolerance = normalized_distance_tolerance
        self.normalized_angle_tolerance = normalized_angle_tolerance
        if additional_conditions is None:
            self.additional_conditions = [self.AC.NONE, self.AC.ONLY_ACB]
        else:
            self.additional_conditions = additional_conditions
        self.valences = valences
        self.maximum_distance_factor = maximum_distance_factor
        self.minimum_angle_factor = minimum_angle_factor
        if isites is None:
            indices = list(range(len(structure)))
        else:
            indices = isites
        self.structure = structure
        logging.info('Setting Voronoi list')
        if voronoi_list2 is not None:
            self.voronoi_list2 = voronoi_list2
        else:
            self.setup_voronoi_list(indices=indices, voronoi_cutoff=voronoi_cutoff)
        logging.info('Setting neighbors distances and angles')
        t1 = time.clock()
        self.setup_neighbors_distances_and_angles(indices=indices)
        t2 = time.clock()
        logging.info('Neighbors distances and angles set up in {:.2f} seconds'.format(t2-t1))

    def setup_voronoi_list(self, indices, voronoi_cutoff):
        """
        Set up of the voronoi list of neighbours by calling qhull
        :param indices: indices of the sites for which the Voronoi is needed
        :param voronoi_cutoff: Voronoi cutoff for the search of neighbours
        :raise RuntimeError: If an infinite vertex is found in the voronoi construction
        """
        self.voronoi_list2 = [None] * len(self.structure)
        self.voronoi_list_coords = [None] * len(self.structure)
        logging.info('Getting all neighbors in structure')
        struct_neighbors = self.structure.get_all_neighbors(voronoi_cutoff, include_index=True)
        t1 = time.clock()
        logging.info('Setting up Voronoi list :')

        for jj, isite in enumerate(indices):
            logging.info('  - Voronoi analysis for site #{:d} ({:d}/{:d})'.format(isite, jj+1, len(indices)))
            site = self.structure[isite]
            neighbors1 = [(site, 0.0, isite)]
            neighbors1.extend(struct_neighbors[isite])
            distances = [i[1] for i in sorted(neighbors1, key=lambda s: s[1])]
            neighbors = [i[0] for i in sorted(neighbors1, key=lambda s: s[1])]
            qvoronoi_input = [s.coords for s in neighbors]
            voro = Voronoi(points=qvoronoi_input, qhull_options="o Fv")
            all_vertices = voro.vertices

            results2 = []
            maxangle = 0.0
            mindist = 10000.0
            for iridge, ridge_points in enumerate(voro.ridge_points):
                if 0 in ridge_points:
                    ridge_vertices_indices = voro.ridge_vertices[iridge]
                    if -1 in ridge_vertices_indices:
                        raise RuntimeError("This structure is pathological,"
                                           " infinite vertex in the voronoi "
                                           "construction")

                    ridge_point2 = max(ridge_points)
                    facets = [all_vertices[i] for i in ridge_vertices_indices]
                    sa = my_solid_angle(site.coords, facets)
                    maxangle = max([sa, maxangle])

                    mindist = min([mindist, distances[ridge_point2]])
                    for iii, sss in enumerate(self.structure):
                        if neighbors[ridge_point2].is_periodic_image(sss):
                            myindex = iii
                            break
                    results2.append({'site': neighbors[ridge_point2],
                                     'angle': sa,
                                     'distance': distances[ridge_point2],
                                     'index': myindex})
            for dd in results2:
                dd['normalized_angle'] = dd['angle'] / maxangle
                dd['normalized_distance'] = dd['distance'] / mindist
            self.voronoi_list2[isite] = results2
            self.voronoi_list_coords[isite] = np.array([dd['site'].coords for dd in results2])
        t2 = time.clock()
        logging.info('Voronoi list set up in {:.2f} seconds'.format(t2-t1))

    def setup_neighbors_distances_and_angles(self, indices):
        """
        Initializes the angle and distance separations
        :param indices: indices of the sites for which the Voronoi is needed
        """
        self.neighbors_distances = [None] * len(self.structure)
        self.neighbors_normalized_distances = [None] * len(self.structure)
        self.neighbors_angles = [None] * len(self.structure)
        self.neighbors_normalized_angles = [None] * len(self.structure)
        for isite in indices:
            results = self.voronoi_list2[isite]
            if results is None:
                continue
            #Initializes neighbors distances and normalized distances groups
            self.neighbors_distances[isite] = []
            self.neighbors_normalized_distances[isite] = []
            normalized_distances = [nb_dict['normalized_distance'] for nb_dict in results]
            isorted_distances = np.argsort(normalized_distances)
            self.neighbors_normalized_distances[isite].append({'min': normalized_distances[isorted_distances[0]],
                                                               'max': normalized_distances[isorted_distances[0]]})
            self.neighbors_distances[isite].append({'min': results[isorted_distances[0]]['distance'],
                                                    'max': results[isorted_distances[0]]['distance']})
            icurrent = 0
            nb_indices = {int(isorted_distances[0])}
            dnb_indices = {int(isorted_distances[0])}
            for idist in iter(isorted_distances):
                wd = normalized_distances[idist]
                if self.maximum_distance_factor is not None:
                    if wd > self.maximum_distance_factor:
                        self.neighbors_normalized_distances[isite][icurrent]['nb_indices'] = list(nb_indices)
                        self.neighbors_distances[isite][icurrent]['nb_indices'] = list(nb_indices)
                        self.neighbors_normalized_distances[isite][icurrent]['dnb_indices'] = list(dnb_indices)
                        self.neighbors_distances[isite][icurrent]['dnb_indices'] = list(dnb_indices)
                        break
                if np.isclose(wd, self.neighbors_normalized_distances[isite][icurrent]['max'],
                              rtol=0.0, atol=self.normalized_distance_tolerance):
                    self.neighbors_normalized_distances[isite][icurrent]['max'] = wd
                    self.neighbors_distances[isite][icurrent]['max'] = results[idist]['distance']
                    dnb_indices.add(int(idist))
                else:
                    self.neighbors_normalized_distances[isite][icurrent]['nb_indices'] = list(nb_indices)
                    self.neighbors_distances[isite][icurrent]['nb_indices'] = list(nb_indices)
                    self.neighbors_normalized_distances[isite][icurrent]['dnb_indices'] = list(dnb_indices)
                    self.neighbors_distances[isite][icurrent]['dnb_indices'] = list(dnb_indices)
                    dnb_indices = {int(idist)}
                    self.neighbors_normalized_distances[isite].append({'min': wd,
                                                                       'max': wd})
                    self.neighbors_distances[isite].append({'min': results[idist]['distance'],
                                                            'max': results[idist]['distance']})
                    icurrent += 1
                nb_indices.add(int(idist))
            else:
                self.neighbors_normalized_distances[isite][icurrent]['nb_indices'] = list(nb_indices)
                self.neighbors_distances[isite][icurrent]['nb_indices'] = list(nb_indices)
                self.neighbors_normalized_distances[isite][icurrent]['dnb_indices'] = list(dnb_indices)
                self.neighbors_distances[isite][icurrent]['dnb_indices'] = list(dnb_indices)
            for idist in range(len(self.neighbors_distances[isite]) - 1):
                dist_dict = self.neighbors_distances[isite][idist]
                dist_dict_next = self.neighbors_distances[isite][idist+1]
                dist_dict['next'] = dist_dict_next['min']
                ndist_dict = self.neighbors_normalized_distances[isite][idist]
                ndist_dict_next = self.neighbors_normalized_distances[isite][idist + 1]
                ndist_dict['next'] = ndist_dict_next['min']
            if self.maximum_distance_factor is not None:
                dfact = self.maximum_distance_factor
            else:
                dfact = self.default_voronoi_cutoff / self.neighbors_distances[isite][0]['min']
            self.neighbors_normalized_distances[isite][-1]['next'] = dfact
            self.neighbors_distances[isite][-1]['next'] = dfact * self.neighbors_distances[isite][0]['min']
            #Initializes neighbors angles and normalized angles groups
            self.neighbors_angles[isite] = []
            self.neighbors_normalized_angles[isite] = []
            normalized_angles = [nb_dict['normalized_angle'] for nb_dict in results]
            isorted_angles = np.argsort(normalized_angles)[::-1]
            self.neighbors_normalized_angles[isite].append({'max': normalized_angles[isorted_angles[0]],
                                                            'min': normalized_angles[isorted_angles[0]]})
            self.neighbors_angles[isite].append({'max': results[isorted_angles[0]]['angle'],
                                                 'min': results[isorted_angles[0]]['angle']})
            icurrent = 0
            nb_indices = {int(isorted_angles[0])}
            dnb_indices = {int(isorted_angles[0])}
            for iang in iter(isorted_angles):
                wa = normalized_angles[iang]
                if self.minimum_angle_factor is not None:
                    if wa < self.minimum_angle_factor:
                        self.neighbors_normalized_angles[isite][icurrent]['nb_indices'] = list(nb_indices)
                        self.neighbors_angles[isite][icurrent]['nb_indices'] = list(nb_indices)
                        self.neighbors_normalized_angles[isite][icurrent]['dnb_indices'] = list(dnb_indices)
                        self.neighbors_angles[isite][icurrent]['dnb_indices'] = list(dnb_indices)
                        break
                if np.isclose(wa, self.neighbors_normalized_angles[isite][icurrent]['min'],
                              rtol=0.0, atol=self.normalized_angle_tolerance):
                    self.neighbors_normalized_angles[isite][icurrent]['min'] = wa
                    self.neighbors_angles[isite][icurrent]['min'] = results[iang]['angle']
                    dnb_indices.add(int(iang))
                else:
                    self.neighbors_normalized_angles[isite][icurrent]['nb_indices'] = list(nb_indices)
                    self.neighbors_angles[isite][icurrent]['nb_indices'] = list(nb_indices)
                    self.neighbors_normalized_angles[isite][icurrent]['dnb_indices'] = list(dnb_indices)
                    self.neighbors_angles[isite][icurrent]['dnb_indices'] = list(dnb_indices)
                    dnb_indices = {int(iang)}
                    self.neighbors_normalized_angles[isite].append({'max': wa,
                                                                    'min': wa})
                    self.neighbors_angles[isite].append({'max': results[iang]['angle'],
                                                         'min': results[iang]['angle']})
                    icurrent += 1
                nb_indices.add(int(iang))
            else:
                self.neighbors_normalized_angles[isite][icurrent]['nb_indices'] = list(nb_indices)
                self.neighbors_angles[isite][icurrent]['nb_indices'] = list(nb_indices)
                self.neighbors_normalized_angles[isite][icurrent]['dnb_indices'] = list(dnb_indices)
                self.neighbors_angles[isite][icurrent]['dnb_indices'] = list(dnb_indices)
            for iang in range(len(self.neighbors_angles[isite]) - 1):
                ang_dict = self.neighbors_angles[isite][iang]
                ang_dict_next = self.neighbors_angles[isite][iang + 1]
                ang_dict['next'] = ang_dict_next['max']
                nang_dict = self.neighbors_normalized_angles[isite][iang]
                nang_dict_next = self.neighbors_normalized_angles[isite][iang + 1]
                nang_dict['next'] = nang_dict_next['max']
            if self.minimum_angle_factor is not None:
                afact = self.minimum_angle_factor
            else:
                afact = 0.0
            self.neighbors_normalized_angles[isite][-1]['next'] = afact
            self.neighbors_angles[isite][-1]['next'] = afact * self.neighbors_angles[isite][0]['max']

    def _precompute_additional_conditions(self, ivoronoi, voronoi, valences):
        additional_conditions = {ac: [] for ac in self.additional_conditions}
        for ips, (ps, vals) in enumerate(voronoi):
            for ac in self.additional_conditions:
                additional_conditions[ac].append(self.AC.check_condition(condition=ac, structure=self.structure,
                                                                         parameters={'valences': valences,
                                                                                     'neighbor_index': vals['index'],
                                                                                     'site_index': ivoronoi}))
        return additional_conditions

    def _precompute_distance_conditions(self, ivoronoi, voronoi):
        distance_conditions = []
        for idp, dp_dict in enumerate(self.neighbors_normalized_distances[ivoronoi]):
            distance_conditions.append([])
            dp = dp_dict['max']
            for ips, (ps, vals) in enumerate(voronoi):
                distance_conditions[idp].append(vals['normalized_distance'] <= dp or
                                                np.isclose(vals['normalized_distance'], dp,
                                                           rtol=0.0, atol=self.normalized_distance_tolerance/2.0))
        return distance_conditions

    def _precompute_angle_conditions(self, ivoronoi, voronoi):
        angle_conditions = []
        for iap, ap_dict in enumerate(self.neighbors_normalized_angles[ivoronoi]):
            angle_conditions.append([])
            ap = ap_dict['max']
            for ips, (ps, vals) in enumerate(voronoi):
                angle_conditions[iap].append(vals['normalized_angle'] >= ap or
                                             np.isclose(vals['normalized_angle'], ap,
                                                        rtol=0.0, atol=self.normalized_angle_tolerance/2.0))
        return angle_conditions

    def neighbors_map(self, isite, distfactor, angfactor, additional_condition):
        if self.neighbors_normalized_distances[isite] is None:
            return None
        dist_where = np.argwhere(np.array([wd['min'] for wd in self.neighbors_normalized_distances[isite]]) <= distfactor)
        if len(dist_where) == 0:
            return None
        idist = dist_where[-1][0]
        ang_where = np.argwhere(np.array([wa['max'] for wa in self.neighbors_normalized_angles[isite]]) >= angfactor)
        if len(ang_where) == 0:
            return None
        iang = ang_where[0][0]
        if self.additional_conditions.count(additional_condition) != 1:
            return None
        i_additional_condition = self.additional_conditions.index(additional_condition)
        return {'i_distfactor': idist, 'i_angfactor': iang, 'i_additional_condition': i_additional_condition}

    def neighbors_surfaces(self, isite, surface_calculation_type=None, max_dist=2.0):
        if self.voronoi_list2[isite] is None:
            return None
        bounds_and_limits = self.voronoi_parameters_bounds_and_limits(isite, surface_calculation_type, max_dist)
        distance_bounds = bounds_and_limits['distance_bounds']
        angle_bounds = bounds_and_limits['angle_bounds']
        surfaces = np.zeros((len(distance_bounds), len(angle_bounds)), np.float)
        for idp in range(len(distance_bounds) - 1):
            this_dist_plateau = distance_bounds[idp + 1] - distance_bounds[idp]
            for iap in range(len(angle_bounds) - 1):
                this_ang_plateau = angle_bounds[iap + 1] - angle_bounds[iap]
                surfaces[idp][iap] = np.absolute(this_dist_plateau*this_ang_plateau)
        return surfaces

    def neighbors_surfaces_bounded(self, isite, surface_calculation_options=None):
        if self.voronoi_list2[isite] is None:
            return None
        if surface_calculation_options is None:
            surface_calculation_options = {'type': 'standard_elliptic',
                                           'distance_bounds': {'lower': 1.2, 'upper': 1.8},
                                           'angle_bounds': {'lower': 0.1, 'upper': 0.8}}
        if surface_calculation_options['type'] in ['standard_elliptic', 'standard_diamond', 'standard_spline']:
            plot_type = {'distance_parameter': ('initial_normalized', None),
                         'angle_parameter': ('initial_normalized', None)}
        else:
            raise ValueError('Type "{}" for the surface calculation in DetailedVoronoiContainer '
                             'is invalid'.format(surface_calculation_options['type']))
        max_dist = surface_calculation_options['distance_bounds']['upper'] + 0.1
        bounds_and_limits = self.voronoi_parameters_bounds_and_limits(isite=isite,
                                                                      plot_type=plot_type,
                                                                      max_dist=max_dist)

        distance_bounds = bounds_and_limits['distance_bounds']
        angle_bounds = bounds_and_limits['angle_bounds']
        lower_and_upper_functions = get_lower_and_upper_f(surface_calculation_options=surface_calculation_options)
        mindist = surface_calculation_options['distance_bounds']['lower']
        maxdist = surface_calculation_options['distance_bounds']['upper']
        minang = surface_calculation_options['angle_bounds']['lower']
        maxang = surface_calculation_options['angle_bounds']['upper']

        f_lower = lower_and_upper_functions['lower']
        f_upper = lower_and_upper_functions['upper']
        surfaces = np.zeros((len(distance_bounds), len(angle_bounds)), np.float)
        for idp in range(len(distance_bounds) - 1):
            dp1 = distance_bounds[idp]
            dp2 = distance_bounds[idp+1]
            if dp2 < mindist or dp1 > maxdist:
                continue
            if dp1 < mindist:
                d1 = mindist
            else:
                d1 = dp1
            if dp2 > maxdist:
                d2 = maxdist
            else:
                d2 = dp2
            for iap in range(len(angle_bounds) - 1):
                ap1 = angle_bounds[iap]
                ap2 = angle_bounds[iap+1]
                if ap1 > ap2:
                    ap1 = angle_bounds[iap + 1]
                    ap2 = angle_bounds[iap]
                if ap2 < minang or ap1 > maxang:
                    continue
                intersection, interror = rectangle_surface_intersection(rectangle=((d1, d2),
                                                                                   (ap1, ap2)),
                                                                        f_lower=f_lower,
                                                                        f_upper=f_upper,
                                                                        bounds_lower=[mindist, maxdist],
                                                                        bounds_upper=[mindist, maxdist],
                                                                        check=False)
                surfaces[idp][iap] = intersection
        return surfaces

    @staticmethod
    def _get_vertices_dist_ang_indices(parameter_indices_list):
        pp0 = [pp[0] for pp in parameter_indices_list]
        pp1 = [pp[1] for pp in parameter_indices_list]
        min_idist = min(pp0)
        min_iang = min(pp1)
        max_idist = max(pp0)
        max_iang = max(pp1)
        i_min_angs = np.argwhere(np.array(pp1) == min_iang)
        i_max_dists = np.argwhere(np.array(pp0) == max_idist)
        pp0_at_min_iang = [pp0[ii[0]] for ii in i_min_angs]
        pp1_at_max_idist = [pp1[ii[0]] for ii in i_max_dists]
        max_idist_at_min_iang = max(pp0_at_min_iang)
        min_iang_at_max_idist = min(pp1_at_max_idist)

        p1 = (min_idist, min_iang)
        p2 = (max_idist_at_min_iang, min_iang)
        p3 = (max_idist_at_min_iang, min_iang_at_max_idist)
        p4 = (max_idist, min_iang_at_max_idist)
        p5 = (max_idist, max_iang)
        p6 = (min_idist, max_iang)

        return [p1, p2, p3, p4, p5, p6]

    def maps_and_surfaces(self, isite, surface_calculation_type=None, max_dist=2.0, additional_conditions=None):
        if self.voronoi_list2[isite] is None:
            return None
        if additional_conditions is None:
            additional_conditions = [self.AC.ONLY_ACB]
        surfaces = self.neighbors_surfaces(isite=isite, surface_calculation_type=surface_calculation_type,
                                           max_dist=max_dist)
        maps_and_surfaces = []
        for cn, value in self._unique_coordinated_neighbors_parameters_indices[isite].items():
            for imap, list_parameters_indices in enumerate(value):
                thissurf = 0.0
                for (idp, iap, iacb) in list_parameters_indices:
                    if iacb in additional_conditions:
                        thissurf += surfaces[idp, iap]
                maps_and_surfaces.append({'map': (cn, imap), 'surface': thissurf,
                                          'parameters_indices': list_parameters_indices})
        return maps_and_surfaces

    def maps_and_surfaces_bounded(self, isite, surface_calculation_options=None, additional_conditions=None):
        if self.voronoi_list2[isite] is None:
            return None
        if additional_conditions is None:
            additional_conditions = [self.AC.ONLY_ACB]
        surfaces = self.neighbors_surfaces_bounded(isite=isite, surface_calculation_options=surface_calculation_options)
        maps_and_surfaces = []
        for cn, value in self._unique_coordinated_neighbors_parameters_indices[isite].items():
            for imap, list_parameters_indices in enumerate(value):
                thissurf = 0.0
                for (idp, iap, iacb) in list_parameters_indices:
                    if iacb in additional_conditions:
                        thissurf += surfaces[idp, iap]
                maps_and_surfaces.append({'map': (cn, imap), 'surface': thissurf,
                                          'parameters_indices': list_parameters_indices})
        return maps_and_surfaces

    def neighbors(self, isite, distfactor, angfactor, additional_condition=None):
        idist = None
        dfact = None
        for iwd, wd in enumerate(self.neighbors_normalized_distances[isite]):
            if distfactor >= wd['min']:
                idist = iwd
                dfact = wd['max']
            else:
                break
        iang = None
        afact = None
        for iwa, wa in enumerate(self.neighbors_normalized_angles[isite]):
            if angfactor <= wa['max']:
                iang = iwa
                afact = wa['min']
            else:
                break
        if idist is None or iang is None:
            raise ValueError('Distance or angle parameter not found ...')

        return [nb for nb in self.voronoi_list2[isite] if
                nb['normalized_distance'] <= dfact and nb['normalized_angle'] >= afact]

    def voronoi_parameters_bounds_and_limits(self, isite, plot_type, max_dist):
        #Initializes the distance and angle parameters
        if self.voronoi_list2[isite] is None:
            return None
        if plot_type is None:
            plot_type = {'distance_parameter': ('initial_inverse_opposite', None),
                         'angle_parameter': ('initial_opposite', None)}
        dd = [dist['min'] for dist in self.neighbors_normalized_distances[isite]]
        dd[0] = 1.0
        if plot_type['distance_parameter'][0] == 'initial_normalized':
            dd.append(max_dist)
            distance_bounds = np.array(dd)
            dist_limits = [1.0, max_dist]
        elif plot_type['distance_parameter'][0] == 'initial_inverse_opposite':
            ddinv = [1.0 / dist for dist in dd]
            ddinv.append(0.0)
            distance_bounds = np.array([1.0 - invdist for invdist in ddinv])
            dist_limits = [0.0, 1.0]
        elif plot_type['distance_parameter'][0] == 'initial_inverse3_opposite':
            ddinv = [1.0 / dist**3.0 for dist in dd]
            ddinv.append(0.0)
            distance_bounds = np.array([1.0 - invdist for invdist in ddinv])
            dist_limits = [0.0, 1.0]
        else:
            raise NotImplementedError('Plotting type "{}" '
                                      'for the distance is not implemented'.format(plot_type['distance_parameter']))
        if plot_type['angle_parameter'][0] == 'initial_normalized':
            aa = [0.0]
            aa.extend([ang['max'] for ang in self.neighbors_normalized_angles[isite]])
            angle_bounds = np.array(aa)
        elif plot_type['angle_parameter'][0] == 'initial_opposite':
            aa = [0.0]
            aa.extend([ang['max'] for ang in self.neighbors_normalized_angles[isite]])
            aa = [1.0 - ang for ang in aa]
            angle_bounds = np.array(aa)
        else:
            raise NotImplementedError('Plotting type "{}" '
                                      'for the angle is not implemented'.format(plot_type['angle_parameter']))
        ang_limits = [0.0, 1.0]
        return {'distance_bounds': distance_bounds, 'distance_limits': dist_limits,
                'angle_bounds': angle_bounds, 'angle_limits': ang_limits}

    def is_close_to(self, other, rtol=0.0, atol=1e-8):
        isclose = (np.isclose(self.normalized_angle_tolerance, other.normalized_angle_tolerance,
                              rtol=rtol, atol=atol) and
                   np.isclose(self.normalized_distance_tolerance, other.normalized_distance_tolerance,
                              rtol=rtol, atol=atol) and
                   self.additional_conditions == other.additional_conditions and
                   self.valences == other.valences)
        if not isclose:
            return isclose
        for isite, site_voronoi in enumerate(self.voronoi_list2):
            self_to_other_nbs = {}
            for inb, nb in enumerate(site_voronoi):
                if nb is None:
                    if other.voronoi_list2[isite] is None:
                        continue
                    else:
                        return False
                else:
                    if other.voronoi_list2[isite] is None:
                        return False
                nb_other = None
                for inb2, nb2 in enumerate(other.voronoi_list2[isite]):
                    if nb['site'] == nb2['site']:
                        self_to_other_nbs[inb] = inb2
                        nb_other = nb2
                        break
                if nb_other is None:
                    return False
                if not np.isclose(nb['distance'], nb_other['distance'],
                                  rtol=rtol, atol=atol):
                    return False
                if not np.isclose(nb['angle'], nb_other['angle'],
                                  rtol=rtol, atol=atol):
                    return False
                if not np.isclose(nb['normalized_distance'], nb_other['normalized_distance'],
                                  rtol=rtol, atol=atol):
                    return False
                if not np.isclose(nb['normalized_angle'], nb_other['normalized_angle'],
                                  rtol=rtol, atol=atol):
                    return False
                if nb['index'] != nb_other['index']:
                    return False
                if nb['site'] != nb_other['site']:
                    return False
        return True

    def get_rdf_figure(self, isite, normalized=True, figsize=None,
                       step_function=None):
        def dp_func(dp):
            return 1.0 - 1.0 / np.power(dp, 3.0)
        import matplotlib.pyplot as plt

        if step_function is None:
            step_function = {'type': 'normal_cdf', 'scale': 0.0001}

        # Initializes the figure
        if figsize is None:
            fig = plt.figure()
        else:
            fig = plt.figure(figsize=figsize)
        subplot = fig.add_subplot(111)
        if normalized:
            dists = self.neighbors_normalized_distances[isite]
        else:
            dists = self.neighbors_distances[isite]

        if step_function['type'] == 'step_function':
            isorted = np.argsort([dd['min'] for dd in dists])
            sorted_dists = [dists[ii]['min'] for ii in isorted]
            dnb_dists = [len(dists[ii]['dnb_indices']) for ii in isorted]
            xx = [0.0]
            yy = [0.0]
            for idist, dist in enumerate(sorted_dists):
                xx.append(dist)
                xx.append(dist)
                yy.append(yy[-1])
                yy.append(yy[-1]+dnb_dists[idist])
            xx.append(1.1*xx[-1])
            yy.append(yy[-1])
        elif step_function['type'] == 'normal_cdf':
            scale = step_function['scale']
            mydists = [dp_func(dd['min']) for dd in dists]
            mydcns = [len(dd['dnb_indices']) for dd in dists]
            xx = np.linspace(0.0, 1.1*max(mydists), num=500)
            yy = np.zeros_like(xx)
            for idist, dist in enumerate(mydists):
                yy += mydcns[idist] * normal_cdf_step(xx, mean=dist, scale=scale)
        else:
            raise ValueError('Step function of type "{}" is not allowed'.format(step_function['type']))
        subplot.plot(xx, yy)

        return fig

    def get_sadf_figure(self, isite, normalized=True, figsize=None,
                       step_function=None):
        def ap_func(ap):
            return np.power(ap, -0.1)
        import matplotlib.pyplot as plt

        if step_function is None:
            step_function = {'type': 'step_function', 'scale': 0.0001}

        # Initializes the figure
        if figsize is None:
            fig = plt.figure()
        else:
            fig = plt.figure(figsize=figsize)
        subplot = fig.add_subplot(111)
        if normalized:
            angs = self.neighbors_normalized_angles[isite]
        else:
            angs = self.neighbors_angles[isite]

        if step_function['type'] == 'step_function':
            isorted = np.argsort([ap_func(aa['min']) for aa in angs])
            sorted_angs = [ap_func(angs[ii]['min']) for ii in isorted]
            dnb_angs = [len(angs[ii]['dnb_indices']) for ii in isorted]
            xx = [0.0]
            yy = [0.0]
            for iang, ang in enumerate(sorted_angs):
                xx.append(ang)
                xx.append(ang)
                yy.append(yy[-1])
                yy.append(yy[-1]+dnb_angs[iang])
            xx.append(1.1*xx[-1])
            yy.append(yy[-1])
        elif step_function['type'] == 'normal_cdf':
            scale = step_function['scale']
            myangs = [ap_func(aa['min']) for aa in angs]
            mydcns = [len(dd['dnb_indices']) for dd in angs]
            xx = np.linspace(0.0, 1.1*max(myangs), num=500)
            yy = np.zeros_like(xx)
            for iang, ang in enumerate(myangs):
                yy += mydcns[iang] * normal_cdf_step(xx, mean=ang, scale=scale)
        else:
            raise ValueError('Step function of type "{}" is not allowed'.format(step_function['type']))
        subplot.plot(xx, yy)

        return fig

    def __eq__(self, other):
        return (self.normalized_angle_tolerance == other.normalized_angle_tolerance and
                self.normalized_distance_tolerance == other.normalized_distance_tolerance and
                self.additional_conditions == other.additional_conditions and
                self.valences == other.valences and
                self.voronoi_list2 == other.voronoi_list2 and
                self.structure == other.structure)

    def __ne__(self, other):
        return not self == other

    def to_bson_voronoi_list2(self):
        """
        Transforms the voronoi_list into a vlist + bson_nb_voro_list, that are BSON-encodable.
        :return: [vlist, bson_nb_voro_list], to be used in the as_dict method
        """
        bson_nb_voro_list2 = [None] * len(self.voronoi_list2)
        for ivoro, voro in enumerate(self.voronoi_list2):
            if voro is None or voro == 'None':
                continue
            site_voro = []
            # {'site': neighbors[nn[1]],
            #  'angle': sa,
            #  'distance': distances[nn[1]],
            #  'index': myindex}
            for nb_dict in voro:
                site = nb_dict['site']
                site_dict = {key: val for key, val in nb_dict.items() if key not in ['site']}
                #site_voro.append([ps.as_dict(), dd]) [float(c) for c in self.frac_coords]
                diff = site.frac_coords - self.structure[nb_dict['index']].frac_coords
                site_voro.append([[nb_dict['index'], [float(c) for c in diff]],
                                  site_dict])
            bson_nb_voro_list2[ivoro] = site_voro
        return bson_nb_voro_list2

    def as_dict(self):
        """
        Bson-serializable dict representation of the VoronoiContainer.
        :return: dictionary that is BSON-encodable
        """
        bson_nb_voro_list2 = self.to_bson_voronoi_list2()
        return {"@module": self.__class__.__module__,
                "@class": self.__class__.__name__,
                "bson_nb_voro_list2": bson_nb_voro_list2,
                # "neighbors_lists": self.neighbors_lists,
                "structure": self.structure.as_dict(),
                "normalized_angle_tolerance": self.normalized_angle_tolerance,
                "normalized_distance_tolerance": self.normalized_distance_tolerance,
                "additional_conditions": self.additional_conditions,
                "valences": self.valences,
                "maximum_distance_factor": self.maximum_distance_factor,
                "minimum_angle_factor": self.minimum_angle_factor}

    @classmethod
    def from_dict(cls, d):
        """
        Reconstructs the VoronoiContainer object from a dict representation of the VoronoiContainer created using
        the as_dict method.
        :param d: dict representation of the VoronoiContainer object
        :return: VoronoiContainer object
        """
        structure = Structure.from_dict(d['structure'])
        voronoi_list2 = from_bson_voronoi_list2(d['bson_nb_voro_list2'], structure)
        maximum_distance_factor = d['maximum_distance_factor'] if 'maximum_distance_factor' in d else None
        minimum_angle_factor = d['minimum_angle_factor'] if 'minimum_angle_factor' in d else None
        return cls(structure=structure, voronoi_list2=voronoi_list2,
                   # neighbors_lists=neighbors_lists,
                   normalized_angle_tolerance=d['normalized_angle_tolerance'],
                   normalized_distance_tolerance=d['normalized_distance_tolerance'],
                   additional_conditions=d['additional_conditions'],
                   valences=d['valences'],
                   maximum_distance_factor=maximum_distance_factor,
                   minimum_angle_factor=minimum_angle_factor)



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from pymatgen import Structure, Lattice, Specie, DummySpecie
import numpy as np

"""
Classes for reading/writing mcsqs files following the rndstr.in format.
"""

__author__ = "Matthew Horton"
__copyright__ = "Copyright 2017, The Materials Project"
__maintainer__ = "Matthew Horton"
__email__ = "mkhorton@lbl.gov"
__status__ = "Production"
__date__ = "October 2017"


class Mcsqs:

    def __init__(self, structure):
        """
        Handle input/output for the crystal definition format
        used by mcsqs and other ATAT codes.

        :param structure: input Structure
        """

        self.structure = structure

    def to_string(self):
        """
        Returns a structure in mcsqs rndstr.in format.
        :return (str):
        """
        
        # add lattice vectors
        m = self.structure.lattice.matrix
        output = ["{:6f} {:6f} {:6f}".format(*m[0]),
                  "{:6f} {:6f} {:6f}".format(*m[1]),
                  "{:6f} {:6f} {:6f}".format(*m[2])]

        # define coord system, use Cartesian
        output.append("1.0 0.0 0.0")
        output.append("0.0 1.0 0.0")
        output.append("0.0 0.0 1.0")
        
        # add species
        for site in self.structure:
            species_str = []
            for sp, occu in sorted(site.species.items()):
                if isinstance(sp, Specie):
                    sp = sp.element
                species_str.append("{}={}".format(sp, occu))
            species_str = ",".join(species_str)
            output.append("{:6f} {:6f} {:6f} {}".format(site.frac_coords[0],
                                                        site.frac_coords[1],
                                                        site.frac_coords[2],
                                                        species_str))

        return "\n".join(output)

    @staticmethod
    def structure_from_string(data):
        """
        Parses a rndstr.in or lat.in file into pymatgen's
        Structure format.

        :param data: contents of a rndstr.in or lat.in file
        :return: Structure object
        """

        data = data.splitlines()
        data = [x.split() for x in data if x]  # remove empty lines

        # following specification/terminology given in manual
        if len(data[0]) == 6:  # lattice parameters
            a, b, c, alpha, beta, gamma = map(float, data[0])
            coord_system = Lattice.from_parameters(a, b, c,
                                                   alpha, beta, gamma).matrix
            lattice_vecs = np.array([
                [data[1][0], data[1][1], data[1][2]],
                [data[2][0], data[2][1], data[2][2]],
                [data[3][0], data[3][1], data[3][2]]
            ], dtype=float)
            first_species_line = 4
        else:
            coord_system = np.array([
                [data[0][0], data[0][1], data[0][2]],
                [data[1][0], data[1][1], data[1][2]],
                [data[2][0], data[2][1], data[2][2]]
            ], dtype=float)
            lattice_vecs = np.array([
                [data[3][0], data[3][1], data[3][2]],
                [data[4][0], data[4][1], data[4][2]],
                [data[5][0], data[5][1], data[5][2]]
            ], dtype=float)
            first_species_line = 6

        scaled_matrix = np.matmul(coord_system, lattice_vecs)
        lattice = Lattice(scaled_matrix)

        all_coords = []
        all_species = []
        for l in data[first_species_line:]:

            all_coords.append(np.array([l[0], l[1], l[2]], dtype=float))

            species_strs = "".join(l[3:])  # join multiple strings back together
            species_strs = species_strs.replace(" ", "")  # trim any white space
            species_strs = species_strs.split(",")  # comma-delimited

            species = {}

            for species_str in species_strs:
                species_str = species_str.split('=')
                if len(species_str) == 1:
                    # assume occupancy is 1.0
                    species_str = [species_str[0], 1.0]
                try:
                    species[Specie(species_str[0])] = float(species_str[1])
                except:
                    species[DummySpecie(species_str[0])] = float(species_str[1])

            all_species.append(species)

        return Structure(lattice, all_species, all_coords)
